# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains functions used for initiation of paths.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_initial_path_kick (:py:func:`.generate_initial_path_kick`)
    Function for generating an initial path by repeatedly kicking a
    phase point.

initiate_kick (:py:func:`.initiate_kick`)
    Helper method, selects either :py:func:`.initiate_kicki` or
    :py:func:`.initiate_kick_max`.

initiate_kicki (:py:func:`.initiate_kicki`)
    A method for initiating a path ensemble by repeatedly modifying
    velocities to find the crossing with the interfaces.

initiate_kick_max (:py:func:`.initiate_kick_max`)
    A method similar to py:meth:`.initiate_kick`. Here, if possible,
    we will use points from the previous paths, closest to the target
    interface.

initiate_path_ensemble_kick (:py:func:`.initiate_path_ensemble_kick`)
    Method to initiate a single path ensemble.

fix_path_by_tis (:py:func:`.fix_path_by_tis`)
    Method to fix an initial path that starts and ends at the wrong
    interface via TIS moves.

"""
import copy
import logging
from pyretis.core.common import compute_weight
from pyretis.core.path import paste_paths, Path
from pyretis.core.tis import shoot, time_reversal
from pyretis.inout.screen import print_to_screen
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = [
    'generate_initial_path_kick',
    'initiate_kick',
    'initiate_kicki',
    'initiate_kick_max',
    'initiate_path_ensemble_kick',
    'fix_path_by_tis'
]


def initiate_kick(simulation, settings, cycle):
    """Run the initiation method for several ensembles.

    Please see documentation for :py:func:`.initiate_path_ensemble_kick`.

    """
    init_settings = settings['initial-path']
    start = init_settings.get('kick-from', 'initial').lower()
    if start == 'previous':
        logger.info('Kick-initiate using previous configuration')
        return initiate_kick_max(simulation, settings, cycle)
    if start == 'initial':
        logger.info('Kick-initiate using initial configuration')
        return initiate_kicki(simulation, settings, cycle)
    errtxt = f'Unknown argument {start} for kick-from'
    logger.error(errtxt)
    raise ValueError(errtxt)


def initiate_kicki(simulation, settings, cycle):
    """Run the initialization for several ensembles.

    Please see documentation for :py:func:`.initiate_path_ensemble_kick`.

    """
    for ensemble, set_ens in zip(simulation.ensembles, settings['ensemble']):
        ens_name = ensemble['path_ensemble'].ensemble_name
        print_to_screen(f'Running "kick" initiation in: {ens_name}')
        logger.info('Initiating path ensemble: %s',
                    ensemble['path_ensemble'].ensemble_name)
        accept, initial_path, status = initiate_path_ensemble_kick(
            ensemble,
            set_ens['tis'],
            cycle)
        yield accept, initial_path, status, ensemble['path_ensemble']


def initiate_kick_max(simulation, settings, cycle):
    """Run the initiation for several ensembles.

    This method is similar to :py:func:`.initiate_kick`, but here we
    update the initial point for an ensemble to use that of the previous
    path (if this exists).

    Please see documentation for :py:func:`.initiate_path_ensemble_kick`.

    """
    last_paths = []
    for ensemble, set_ens in zip(simulation.ensembles, settings['ensemble']):
        path_ensemble = ensemble['path_ensemble']
        logger.info('Initiating path ensemble %s', path_ensemble.ensemble_name)
        ens_name = path_ensemble.ensemble_name
        print_to_screen(f'Running kick-max initiation in: {ens_name}')
        if last_paths:
            # Look for the phase point closest to the middle interface.
            middle = path_ensemble.interfaces[1]
            current = None
            min_dist = float('inf')
            for last_path in last_paths:
                for phase_point in last_path.phasepoints:
                    dist = middle - phase_point.order[0]
                    if 0 <= dist <= min_dist:
                        min_dist = dist
                        current = phase_point.copy()
            if current is not None:
                logger.info('Initial state set to: %s', current)
                ensemble['system'] = current
        accept, initial_path, status = initiate_path_ensemble_kick(
            ensemble,
            set_ens['tis'],
            cycle)
        last_paths.append(initial_path)
        yield accept, initial_path, status, path_ensemble


def initiate_path_ensemble_kick(ensemble, tis_settings, cycle):
    """Run the "kick" initiate for a given ensemble.

    Parameters
    ----------
    ensemble: dict
        It contains:

        * `path_ensemble` : object like :py:class:`.PathEnsemble`
          The path ensemble to create an initial path for.
        * `system` : object like :py:class:`.System`
          The system is used here since we need access to the temperature
          and to the particle list.
        * `order_function` : object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine` : object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen` : object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.

    tis_settings : dict
        This dictionary contains the TIS settings.
    cycle : integer, optional
        The cycle number we are initiating at, typically this will be 0
        which is the default value.

    Returns
    -------
    out[0] : boolean
        True if the initial path was accepted.
    out[1] : object like py:class:`.PathBase`
        The initial path.
    out[2] : string
        The status of the path.

    """
    initial_path = None
    logger.info('Will generate initial path by kicking')
    ensemble['engine'].exe_dir = \
        ensemble['path_ensemble'].directory['generate']

    for attempt in generate_initial_path_kick(ensemble, tis_settings):
        if attempt[0] is True:
            initial_path = attempt[-1]
            break
    accept = True
    status = 'ACC'
    ensemble['path_ensemble'].add_path_data(initial_path, status, cycle)
    # Ask the engine to do clean up after the initialisation:
    ensemble['engine'].clean_up()

    return accept, initial_path, status


def generate_initial_path_kick(ensemble, tis_settings):
    """Generate an initial path with the kicking method.

    This function will generate an initial path by repeatedly kicking a
    phase-space point until the middle interface is crossed.
    The point before and after kicking are stored, so when the
    middle interface is crossed we have two points we can integrate
    forward and backward in time. This function is intended for use with
    TIS. For use with RETIS one should set the appropriate
    `tis_settings` so that the starting conditions are fine (i.e. for
    the [0^-] ensemble it might be different for the other ensembles).

    Parameters
    ----------
    ensemble: dict.
        It contains:

        * `path_ensemble` : object like :py:class:`.PathEnsemble`
          The path ensemble to create an initial path for.
        * `system` : object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function` : object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine` : object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen` : object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.

    tis_settings : dict
        This dictionary contains settings for TIS. Explicitly used here:

        * `start_cond`: string, starting condition, 'L'eft or 'R'ight.
        * `maxlength`: integer, maximum allowed length of paths.

    Yields
    ------
    out[0] : boolean
        True if we are finished.
    out[1] : string
        The current status of this method.
    out[2] : object like :py:class:`.PathBase` or None
        If we are finished (and successful) this is the generated
        initial path. Otherwise, this is equal to None.

    """
    engine = ensemble['engine']
    interfaces = ensemble['path_ensemble'].interfaces
    rgen = ensemble['path_ensemble'].rgen
    initial_system = ensemble['system'].copy()
    middle = interfaces[1]
    kick_required = True
    if set(ensemble['path_ensemble'].start_condition) == set(['R', 'L']):
        initial_op = initial_system.order[0]
        logger.info("Initial OP: %s", initial_op)

        # If we are allowed to end anywhere and the initial system is between
        # the two interfaces, don't kick.
        if interfaces[0] < initial_op < interfaces[-1]:
            kick_required = False
        # Else kick to the closest interfaces
        elif abs(interfaces[0]-initial_op) < abs(interfaces[-1]-initial_op):
            middle = interfaces[0]
        else:
            middle = interfaces[-1]
    maxlen = tis_settings['maxlength']

    while True:
        # Start from the initial system:
        ensemble['system'] = initial_system.copy()
        ensemble['interfaces'] = interfaces
        if kick_required:
            logger.info('Searching crossing with middle interface')
            left_point, right_point = engine.kick_across_middle(ensemble,
                                                                middle,
                                                                tis_settings)
        else:
            logger.info('Not searching a crossing with the middle interface')
            right_point = initial_system.copy()
            # Set starting point for both propagators to the current system
            left_point = initial_system.copy()

        logger.info('Propagating from crossing points')
        # kick_across_middle will return two points, one immediately
        # left of the interface and one immediately right of the
        # interface. So we have two points (`leftpoint` and the
        # current `system.particles`). We then propagate the current
        # phase point forward:
        logger.info('Propagating forward from crossing points '
                    'for initial path')
        path_forw = Path(rgen=rgen, maxlen=maxlen)
        ensemble['system'] = right_point
        success, msg = engine.propagate(path_forw, ensemble,
                                        reverse=False)

        if 'exp' in tis_settings.get('shooting_move', 'sh'):
            path_forw.status = 'ACC'
            path_forw.generated = ('ki', 0, 0, 0)
            yield True, 'Initial path generated.', path_forw

        if not success:
            logger.warning('Forward path not successful: %s', msg)
            yield False, f'Forward path failed: {msg}', None
            continue
        # And we propagate the `leftpoint` backward:
        path_back = Path(rgen=rgen, maxlen=maxlen)
        ensemble['system'] = left_point
        success, msg = engine.propagate(path_back, ensemble,
                                        reverse=True)
        if not success:
            logger.warning('Backward path not successful: %s', msg)
            yield False, f'Backward path failed: {msg}', None
            continue
        # Merge backward and forward, here we do not set maxlen since
        # both backward and forward may have this length:
        initial_path = paste_paths(path_back, path_forw, overlap=False)
        if initial_path.length >= maxlen:
            logger.warning('Initial path too long (exceeded "MAXLEN")')
            yield False, 'Initial path was too long.', None
            continue
        start, end, _, _ = initial_path.check_interfaces(interfaces)
        # OK, now it is time to check the path:
        # 0) We can start at the starting condition, pass the middle
        #    and continue all the way to the end - perfect!
        # 1) We can start at the starting condition, pass the middle
        #    and return to starting condition - this is perfectly fine.
        # 2) We can start at the wrong interface, pass the middle and
        #    end at the same (wrong) interface - we fix this by
        #    additional TIS moves.
        # 3) We can start at wrong interface and end at the starting
        #    condition. To fix this, we reverse the path.
        if start in set(ensemble['path_ensemble'].start_condition):
            # Case 0) & 1):
            break
        else:
            # Now we do the other cases:
            if end in set(ensemble['path_ensemble'].start_condition):
                # Case 3) (and start != start_cond):
                logger.info('Initial path is in the wrong direction')
                initial_path = initial_path.reverse(ensemble['order_function'])
                logger.info('Initial path has been reversed!')
                break
            elif end == start:
                # Case 2):
                logger.info('Initial path starts and ends at wrong interface')
                logger.info('Will perform TIS moves to fix it!')
                yield False, 'Trying to fix path by TIS moves', None

                initial_path = fix_path_by_tis(initial_path, ensemble,
                                               tis_settings)
                break
            else:
                # This is reached if the start/end conditions
                # has been modified or are inconsistent.
                logger.warning('Could not generate initial path, will retry!')
                yield (False, 'Could not generate initial path will retry!',
                       None)
                continue
    initial_path.status = 'ACC'
    initial_path.generated = ('ki', 0, 0, 0)
    if tis_settings.get('high_accept', False):
        move = tis_settings.get('shooting_move', 'sh')
        if move == 'wf':
            int_cap = tis_settings.get('interface_cap', interfaces[2])
            interfaces = [interfaces[0], interfaces[1], int_cap]
        if move in ('ss', 'wf'):
            initial_path.weight = compute_weight(initial_path,
                                                 interfaces,
                                                 move)
    yield True, 'Initial path generated.', initial_path


def _get_help(start_cond, interfaces):
    """Define some helper methods for :py:func:`.fix_path_by_tis`.

    This method returns two methods that :py:func:`.fix_path_by_tis`
    can use to determine if a new path is an improvement compared to
    the current path and if the "fixing" is done.

    Parameters
    ----------
    start_cond : string
        The starting condition (from the TIS settings). Left or Right.
    interfaces : list of floats
        The interfaces, on form ``[left, middle, right]``.

    Returns
    -------
    out[0] : method
        The method which determines if a new path represents an
        improvement over the current path.
    out[1] : method
        The method which determines if we are done, that is if we
        can accept the current path.

    """
    left, middle, right = interfaces
    improved, done = None, None
    if start_cond == 'R' or middle == right:
        def improved_r(newp, current):
            """Return True if new path is an improvement."""
            return (newp.ordermax[0] > current.ordermax[0] and
                    newp.ordermin[0] < middle)

        def done_r(path):
            """Return True if the path can be accepted."""
            return path.ordermax[0] > right

        improved = improved_r
        done = done_r
    elif start_cond == 'L' or middle == left:
        def improved_l(newp, current):
            """Return True if new path is an improvement."""
            return (newp.ordermin[0] < current.ordermin[0] and
                    newp.ordermax[0] > middle)

        def done_l(path):
            """Return True if the path can be accepted."""
            return path.ordermin[0] < left

        improved = improved_l
        done = done_l
    else:
        logger.error('Unknown start condition (should be "R" or "L")')
        raise ValueError('Unknown start condition (should be "R" or "L")')
    return improved, done


def fix_path_by_tis(initial_path, ensemble, tis_settings):
    """Fix a path that starts and ends at the wrong interfaces.

    The given path is amended by performing TIS moves (shooting and
    time reversal). Note that this function is intended to be used
    in connection with the initialisation.

    Parameters
    ----------
    initial_path : object like :py:class:`.PathBase`
        This is the initial path to fix. It starts & ends at the
        wrong interface.
    ensemble: dict.
        It contains:

        * `path_ensemble` : object like :py:class:`.PathEnsemble`
          The path ensemble to create an initial path for.
        * `system` : object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function` : object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine` : object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen` : object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.

    tis_settings : dict
        Settings for TIS method, here we explicitly use:

        * `start_cond`: string which defines the start condition.
        * `maxlength`: integer which gives the maximum allowed path
          length.

        Note that we here explicitly set some local TIS settings for
        use in the `make_tis_step` function.

    Returns
    -------
    out : object like :py:class:`.PathBase`
        The amended path.

    """
    logger.debug('Attempting to fix path by running TIS moves.')

    path_ensemble = ensemble['path_ensemble']
    local_tis_settings = copy.deepcopy(tis_settings)
    local_tis_settings['allowmaxlength'] = True
    local_tis_settings['aimless'] = True
    improved, check_ok = _get_help(path_ensemble.start_condition,
                                   path_ensemble.interfaces)

    backup_path = True
    path_ok = False

    path_ensemble.last_path = initial_path

    while not path_ok:
        logger.debug('Performing a TIS move to improve the initial path')
        if backup_path:  # Move the initial path to safe place:
            logger.debug('Moving initial_path')
            path_ensemble.move_path_to_generate(initial_path, '_')
            backup_path = False

        accept, trial, _ = time_reversal(initial_path,
                                         ensemble['order_function'],
                                         path_ensemble.interfaces,
                                         path_ensemble.start_condition)

        if not accept:
            target = path_ensemble.interfaces[1]
            best_frame = ensemble['system']

            for phasepoint in initial_path.phasepoints:
                if abs(best_frame.order[0] - target) > \
                        abs(phasepoint.order[0] - target):
                    best_frame = phasepoint.copy()

            accept, trial, _ = shoot(ensemble,
                                     local_tis_settings,
                                     path_ensemble.start_condition,
                                     best_frame)

        if accept:
            if improved(trial, initial_path):
                logger.debug('TIS move improved path.')
                initial_path = trial
                path_ensemble.last_path = initial_path
                backup_path = True
            else:
                logger.debug('TIS move did not improve path')
            path_ok = check_ok(initial_path)

    initial_path.generated = ('ki', 0, 0, 0)
    initial_path.status = 'ACC'
    return initial_path



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the simulationio module."""
from io import StringIO
import logging
import os
import unittest
from unittest.mock import patch
from pyretis.inout.fileio import FileIO
from pyretis.inout.archive import PathStorage
from pyretis.inout.formats import (
    PathIntFormatter,
    ThermoTableFormatter,
)
from pyretis.inout.screen import ScreenOutput
from pyretis.inout.simulationio import (
    get_task_type,
    get_file_mode,
    OutputTask,
    OUTPUT_TASKS,
    Task,
    task_from_settings,
)
from pyretis.engines.external import ExternalMDEngine
from pyretis.engines.internal import Verlet
from pyretis.inout.settings import add_default_settings
from pyretis.simulation import (
    Simulation,
    UmbrellaWindowSimulation,
    SimulationMD,
    SimulationNVE,
    SimulationMDFlux,
    SimulationTIS,
    SimulationRETIS,
)
logging.disable(logging.CRITICAL)


class TestTask(unittest.TestCase):
    """Run the tests for :py:class`.Task`"""

    def test_when_change(self):
        """Test that we can change the "when" property."""
        task = Task(None)
        self.assertIsNone(task.when)
        task.when = None
        self.assertIsNone(task.when)
        task.when = {'every': 1}
        self.assertEqual(task.when, {'every': 1})
        task.when = None
        self.assertIsNone(task.when)
        task.when = {'every': 1, 'all-the-time': 100}
        self.assertEqual(task.when, {'every': 1})

    def test_task_dict(self):
        """Test that the dictionary returned from a task is ok."""
        task = Task(None)
        task.when = {'every': 2}
        task_dict = task.task_dict()
        self.assertEqual(task_dict['when'], {'every': 2})

    def test_execute_now(self):
        """Test the execute now method."""
        step = {'step': 10, 'start': 2, 'stepno': 8}
        task = Task(None)

        task.when = None
        exe = task.execute_now(step)
        self.assertTrue(exe)

        task.when = {'every': 2}
        exe = task.execute_now(step)
        self.assertTrue(exe)

        task.when = {'every': 3}
        exe = task.execute_now(step)
        self.assertFalse(exe)

        task.when = {'at': 10}
        exe = task.execute_now(step)
        self.assertTrue(exe)

        task.when = {'at': (1, 9, 10)}
        exe = task.execute_now(step)
        self.assertTrue(exe)


class DummyExternal(ExternalMDEngine):
    """A dummy external engine. Only useful for testing."""

    def __init__(self, input_path, timestep, subcycles):
        """Initialise the dummy engine."""
        super().__init__('External engine for testing!', timestep,
                         subcycles)
        self.input_path = input_path

    def step(self, system, name):
        """Perform a single dummy step."""

    def _read_configuration(self, filename):
        """Read a xyz configuration."""

    def _reverse_velocities(self, filename, outfile):
        """Reverse velocities with a dummy method."""

    def _extract_frame(self, traj_file, idx, out_file):
        """Extract a frame, dummy method."""

    def _propagate_from(self, name, path, system, order_function, interfaces,
                        msg_file, reverse=False):
        """Propagate with the engine, dummy method."""
        # pylint: disable=too-many-arguments

    def modify_velocities(self, system, rgen, sigma_v=None, aimless=True,
                          momentum=False, rescale=None):
        """Modify velocities, dummy method."""
        # pylint: disable=too-many-arguments


class TestMethods(unittest.TestCase):
    """Run the tests for methods defined in the module."""

    def test_get_task_type(self):
        """Test the get_task_type method."""
        task = {'type': 'pathensemble'}
        task_type = get_task_type(task, None)
        self.assertEqual(task_type, task['type'])
        task = {'type': 'path-traj-{}'}
        task_type = get_task_type(task, None)
        self.assertEqual(task_type, 'path-traj-int')
        engine = DummyExternal('test', 0.0, 1)
        task_type = get_task_type(task, engine)
        self.assertEqual(task_type, 'path-traj-ext')

    def test_get_file_mode(self):
        """Test the get_file_mode method."""
        settings = {'output': {'backup': 'overwrite'}}
        mode = get_file_mode(settings)
        self.assertEqual(mode, 'w')

        settings = {'output': {'backup': 'append'}}
        mode = get_file_mode(settings)
        self.assertEqual(mode, 'a')

        settings = {'output': {'backup': 10}}
        mode = get_file_mode(settings)
        self.assertEqual(mode, 'w')
        self.assertEqual(settings['output']['backup'], 'backup')

    def test_task_from_settings(self):
        """Test that we can create tasks from settings."""
        for engine in (None, DummyExternal('test', 0.1, 1), Verlet(0.1)):
            for key, val in OUTPUT_TASKS.items():
                task_dict = {'type': key, 'name': f'test-{key}'}
                settings = {}
                add_default_settings(settings)
                directory = os.path.join('this', 'is', 'a', 'fake', 'path')
                task = task_from_settings(task_dict, settings, directory,
                                          engine)
                self.assertEqual(task.result, val['result'])
                self.assertEqual(task.target, val['target'])
                self.assertEqual(task.when['every'],
                                 settings['output'][val['when']])
                if val['target'] == 'file':
                    self.assertIsInstance(task.writer, FileIO)
                elif val['target'] == 'screen':
                    self.assertIsInstance(task.writer, ScreenOutput)
                elif val['target'] == 'file-archive':
                    self.assertIsInstance(task.writer, PathStorage)
                if 'formatter' in val:
                    self.assertIsInstance(task.writer.formatter,
                                          val['formatter'])

        task_dict = {'type': 'energy', 'name': 'test-energy'}
        settings = {}
        add_default_settings(settings)
        settings['output']['energy-file'] = 0
        task = task_from_settings(task_dict, settings, '.', None)
        self.assertIsNone(task)

        task_dict = {'type': 'path-traj-{}', 'name': 'test-path-traj'}
        settings = {}
        add_default_settings(settings)
        task = task_from_settings(task_dict, settings, '.', Verlet(0.1))
        self.assertIsInstance(task.writer.formatter, PathIntFormatter)

        task = task_from_settings(task_dict, settings, '.',
                                  DummyExternal('test', 0.1, 1))
        self.assertIsInstance(task.writer, PathStorage)

        task_dict = {'type': 'does-not-exist', 'name': 'test-energy'}
        settings = {}
        add_default_settings(settings)
        with self.assertRaises(KeyError):
            task_from_settings(task_dict, settings, '.', None)

        OUTPUT_TASKS['thermo-web'] = {
            'target': 'web',
            'filename': 'thermo.txt',
            'result': ('thermo',),
            'when': 'energy-file',
            'formatter': ThermoTableFormatter,
        }
        task_dict = {'type': 'thermo-web', 'name': 'test-web'}
        settings = {}
        add_default_settings(settings)
        task = task_from_settings(task_dict, settings, '.', None)
        self.assertIsNone(task)
        task_dict = {'type': 'thermo-screen', 'name': 'thermo-screen'}
        task = task_from_settings(task_dict, settings, '.', None,
                                  progress=False)
        self.assertIsNotNone(task)
        task = task_from_settings(task_dict, settings, '.', None,
                                  progress=True)
        self.assertIsNone(task)

    def test_create_tasks(self):
        """Test that we can create tasks from settings for simulations."""
        settings = {}
        add_default_settings(settings)
        for simulation in (UmbrellaWindowSimulation,
                           SimulationMD, SimulationNVE,
                           SimulationMDFlux, SimulationTIS,
                           SimulationRETIS):
            sim = Simulation({}, {})
            sim.simulation_output = simulation.simulation_output
            sim.create_output_tasks(settings, progress=False)
            for task in sim.output_tasks:
                self.assertIsInstance(task, OutputTask)


class TestOutputTask(unittest.TestCase):
    """Test the functionality of the output task class."""

    def test_task_dict(self):
        """Test the task dict output."""
        formatter = ThermoTableFormatter()
        writer = ScreenOutput(formatter)
        task = OutputTask('thermo-test', ('thermo', ), writer, {'every': 1})
        task_dict = task.task_dict()
        self.assertEqual(task_dict['writer'], ScreenOutput)
        self.assertEqual(task_dict['formatter'], ThermoTableFormatter)

    def test_output(self):
        """Test that we can output from the task."""
        formatter = ThermoTableFormatter()
        writer = ScreenOutput(formatter)
        task = OutputTask('thermo-test', ('thermo', ), writer, {'every': 1})
        result = {
            'cycle': {'step': 110, 'stepno': 10},
            'thermo': {'vpot': 1.0, 'ekin': 2.0,
                       'etot': 3.0, 'temp': 4.0, 'press': 5.0},
        }
        correct = [
            ('#     Step          Temp           Pot           '
             'Kin           Tot         Press'),
            ('       110             4             1             2'
             '             3             5'),
            ('      1000             4             1             2'
             '             3             5'),
        ]
        idx = 0
        with patch('sys.stdout', new=StringIO()) as stdout:
            task.output(result)
            result['cycle']['step'] = 1000
            task.output(result)
            for linei in stdout.getvalue().strip().split('\n'):
                if linei:
                    self.assertEqual(linei, correct[idx])
                    idx += 1
        task = OutputTask('thermo-test', ('thermo', ), writer, {'every': 99})
        out = task.output(result)
        self.assertFalse(out)
        result['cycle'] = {'step': 99, 'stepno': 99}
        del result['thermo']
        out = task.output(result)
        self.assertFalse(out)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains a class for a generic property."""
import numpy as np

__all__ = ['Property']

np.set_printoptions(legacy='1.21')


class Property:
    """A generic numerical value with standard deviation and average.

    A generic object to store values obtained during a simulation.
    It will maintain the mean and variance as values are added using
    Property.add(val)

    Attributes
    ----------
    desc : string
        Description of the property.
    nval : integer
        Number values stored.
    mean : float
        The current mean.
    delta2 : float
        Helper variable used for calculating the variance.
    variance : float
        The current variance
    val : list
        Store all values.

    Parameters
    ----------
    desc : string, optional
        Used to set the attribute desc.

    Examples
    --------
    >>> from pyretis.core.properties import Property
    >>> ener = Property(desc='Energy of the particle(s)')
    >>> ener.add(42.0)
    >>> ener.add(12.220)
    >>> ener.add(99.22)
    >>> ener.mean

    """

    def __init__(self, desc=''):
        """Initialise the property.

        Parameters
        ----------
        desc : string, optional
            Description of the object.

        """
        self.desc = desc
        self.nval = 0
        self.mean = 0.0
        self.delta2 = 0.0
        self.variance = 0.0
        self.val = []

    def add(self, val):
        """Add a value to the property & update the mean and variance.

        Parameters
        ----------
        val : float or another type (list, numpy.array)
            The value to add.

        Returns
        -------
        out : None
            Returns `None` but updates the mean and variance.

        """
        self.nval += 1
        self.val.append(val)
        self.update_mean_and_variance()

    def update_mean_and_variance(self):
        """Calculate the mean and variance on the fly.

        Source:
        http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance

        Returns
        -------
        out : None
            Returns `None` but updates the mean and variance.

        Note
        ----
        Consider if this should be moved/deleted and just
        replaced with a function from the analysis package.

        """
        val = self.val[-1]  # most recent value
        delta = val - self.mean
        self.mean += delta/float(self.nval)
        self.delta2 += delta * (val - self.mean)
        if self.nval < 2:
            self.variance = float('inf')
        else:
            self.variance = self.delta2/float(self.nval - 1)

    def dump_to_file(self, filename):
        """Dump the contents in `self.val` to a file.

        Parameters
        ----------
        filename : string
            Name/path of file to write.

        Note
        ----
        Consider if this should be moved/deleted and just
        replaced with a function from a more general input-output
        module

        """
        np.savetxt(filename, self.val)

    def __str__(self):
        """Return string description of the property."""
        return self.desc



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for the table formatters."""
import logging
import os
import tempfile
import unittest
import numpy as np
from pyretis.core.pathensemble import PathEnsemble
from pyretis.inout.formats.txt_table import (
    TxtTableFormatter,
    PathTableFormatter,
    ThermoTableFormatter,
    RETISResultFormatter,
    txt_save_columns,
)
from .help import turn_on_logging


logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


class TestTxtTableFormatter(unittest.TestCase):
    """Test that the table formatter works as intended."""

    def test_table_formatter(self):
        """Test the table formatter."""
        table = {
            'title': 'Test table formatter.',
            'var': ['step', 'temp', 'pot', 'ekin'],
            'format': {
                'labels': ['Step', 'Temperature', 'Potential energy',
                           'Kinetic energy'],
                'width': (10, 12),
                'spacing': 2,
                'row_fmt': ['{:> 10d}', '{:> 12.6g}'],
                }
        }
        formatter = TxtTableFormatter(table['var'], table['title'],
                                      **table['format'])
        rows = [
            {'temp': 300, 'pot': 1.0, 'ekin': 2.0},
            {'temp': 300.1, 'pot': 2.1, 'ekin': 3.1},
        ]
        self.assertEqual(
            formatter.header,
            '#     Step   Temperature  Potential energy  Kinetic energy',
        )
        correct = [
            '         1           300             1             2',
            '         2         300.1           2.1           3.1',
        ]
        for i, row in enumerate(rows):
            for line in formatter.format(i+1, row):
                self.assertEqual(line, correct[i])

    def test_table_formatter_missing(self):
        """Test the table formatter when some input is missing."""
        table = {
            'title': 'Test table formatter.',
            'var': ['step', 'temp', 'pot', 'ekin'],
            'format': {
                'labels': ['Step', 'Temperature', 'Potential energy',
                           'Kinetic energy'],
                }
        }
        formatter = TxtTableFormatter(table['var'], table['title'],
                                      **table['format'])
        rows = [
            {'temp': 500, 'pot': 5.0, 'ekin': 20.0},
            {'temp': 500.1, 'pot': 5.1, 'ekin': 30.1},
        ]
        self.assertEqual(
            formatter.header,
            '#       Step  Temperature Potential energy Kinetic energy'
        )
        correct = [
            '           1          500                5             20',
            '           2        500.1              5.1           30.1',
        ]
        for i, row in enumerate(rows):
            for line in formatter.format(i+1, row):
                self.assertEqual(line, correct[i])

    def test_table_formatter_short(self):
        """Test the table formatter when some labels are short."""
        table = {
            'title': 'Test table formatter.',
            'var': ['step', 'x', 'y', 'z'],
            'format': {
                'labels': ['Step', 'x', 'y', 'z'],
                'width': (10, 6),
                }
        }
        formatter = TxtTableFormatter(table['var'], table['title'],
                                      **table['format'])
        rows = [
            {'x': 1, 'y': 2, 'z': -1},
            {'x': 3.14, 'y': -2.71, 'z': -100.1},
        ]
        self.assertEqual(
            formatter.header,
            '#     Step      x      y      z',
        )
        correct = [
            '         1      1      2     -1',
            '         2   3.14  -2.71 -100.1',
        ]
        for i, row in enumerate(rows):
            for line in formatter.format(i+1, row):
                self.assertEqual(line, correct[i])


class TableFMTTest(unittest.TestCase):
    """Test that table writers work as intended."""

    def test_thermo_table(self):
        """Test the thermo table."""
        table = ThermoTableFormatter()
        data = dict(step=100, temp=1.2345, vpot=5.4321e3, ekin=2.222,
                    etot=3.456, press=1.011e9)
        line = ('       100        1.2345        5432.1         '
                '2.222         3.456     1.011e+09')
        for lines in table.format(100, data):
            self.assertMultiLineEqual(lines, line)
            break
        data = dict(step=100, temp=1.2345, vpot=5.4321e3, ekin=2.222,
                    etot=3.456, press=101.11111111)
        line = ('       100        1.2345        5432.1         '
                '2.222         3.456       101.111')
        for lines in table.format(100, data):
            self.assertMultiLineEqual(lines, line)
            break

    def test_path_table(self):
        """Test the path data table."""
        table = PathTableFormatter()
        ens = PathEnsemble(1, [-0.9, -0.9, 1.0])
        ens.nstats['npath'] = 10
        ens.nstats['nshoot'] = 8
        ens.nstats['ACC'] = 4
        ens.nstats['BWI'] = 3
        ens.nstats['FTL'] = 2
        ens.nstats['FTX'] = 1
        correct = ('       101             4             3             0'
                   '             2             0             1')
        for line in table.format(101, ens):
            self.assertEqual(line, correct)
            break

    def test_column_writer(self):
        """Test the column writer method."""
        with tempfile.NamedTemporaryFile() as temp:
            columns = [np.random.rand(10) for i in range(5)]
            txt_save_columns(temp.name, 'Test data', columns, backup=True)
            temp.flush()
        with tempfile.NamedTemporaryFile() as temp:
            columns = [np.random.rand(10-i) for i in range(5)]
            with turn_on_logging():
                with self.assertLogs('pyretis.inout.formats.txt_table',
                                     level='WARNING'):
                    txt_save_columns(temp.name, 'Test data', columns,
                                     backup=True)
            temp.flush()

    def test_retis_formatter(self):
        """Test the formatter for RETIS results."""

        formatter = RETISResultFormatter()
        ens = PathEnsemble(1, [-0.9, -0.9, 1.0])
        ens.nstats['npath'] = 10
        ens.nstats['nshoot'] = 8
        ens.nstats['ACC'] = 4
        ens.nstats['BWI'] = 3
        ens.nstats['FTL'] = 2
        ens.nstats['FTX'] = 1
        ens.paths = [
            {
                'generated': ('s-', 0, 0, 0), 'status': 'NCR', 'length': 69,
                'ordermax': (-0.8977260384533993, 32),
                'ordermin': (-0.9001050210610345, 0),
                'interface': ('L', '*', 'L'), 'cycle': 101}
        ]
        correct = [
            '# Results for path ensemble [0^+] at cycle 101:',
            ('# Generated path with status "NCR", move "swap from -" '
             'and length 69.'),
            '# Order parameter max was: -0.8977260384533993 at index 32.',
            '# Order parameter min was: -0.9001050210610345 at index 0.',
            '# Path ensemble statistics:',
            ('# Ensemble     Cycle    Accepted         BWI         BTL'
             '         FTL         BTX         FTX'),
            ('     [0^+]       101           4           3           0'
             '           2           0           1'),
            '\n',
        ]
        for linei, linej in zip(formatter.format(101, ens), correct):
            self.assertEqual(linei, linej)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two simulations."""
import argparse
import os
import sys
import colorama
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.core.pathensemble import generate_ensemble_name
from pyretis.testing.compare import (
    compare_files,
    compare_files_columns,
    compare_pathensemble_files,
    compare_traj_archive,
)


TRAJ = 'traj'
RETIS_RST = 'retis.rst'


def soft_archive_comparison(dir1, dir2):
    """Compare two archives. Here we accept a difference in energies.

    For GROMACS1 and 2, the potential energy output is different.
    This is due to a difference in the way GROMACS add dispersion
    corrections when continuing a simulation. So, here we expect
    a failure for the energy files. We check these energy files
    manually, by skipping the potential energy terms.

    Parameters
    ----------
    dir1 : string
        The path to the first archive used in the comparison.
    dir2 : string
        The path to the second archive used in the comparison.

    Returns
    -------
    out : tuple of strings
        These are the files which were found to be different.

    """
    errors = []
    for (file1, file2) in compare_traj_archive(dir1, dir2):
        equal, _ = compare_files_columns(
            file1, file2, file_type='energy', skip=['vpot'],
        )
        if not equal:
            errors.append(file1, file2)
    return errors


def compare_ensemble(run1, run2, ensemble,
                     path_skip=None, energy_skip=None, traj_skip=False):
    """Run the comparison for an ensemble.

    Parameters
    ----------
    run1 : string
        The path to the first directory to use for the comparison.
    run2 : string
        The path to the second directory to use for the comparison.
    ensemble : string
        The label for the current ensemble we are investigating, e.g.
        000, 001, or similar.
    path_skip : list of integers, optional
        This selects columns to skip when comparing path ensemble
        data. For instance, when loading trajectories, we essentially
        start a new simulation and not all columns in the path ensemble
        files can then be compared.
    energy_skip : list of strings, optional
        This selects columns to skip in the energy file. This is useful
        when comparing potential energies from GROMACS runs, where these
        don't match due to the way GROMACS handles dispersion
        corrections.

    """
    print_to_screen(f'Comparing for "{ensemble}"', level='info')
    for filei in ('energy.txt', 'order.txt', 'pathensemble.txt'):
        print_to_screen(f'\tComparing {filei} files...')
        file1 = os.path.join(run1, ensemble, filei)
        file2 = os.path.join(run2, ensemble, filei)
        if filei == 'pathensemble.txt':
            equal, msg = compare_pathensemble_files(
                file1, file2, skip=path_skip
            )
        elif filei == 'energy.txt' and energy_skip:
            equal, msg = compare_files_columns(
                file1, file2, file_type='energy', skip=energy_skip,
            )
        else:
            equal, msg = compare_files(
                file1, file2, skip=None, mode='numerical'
            )
        lvl = 'success' if equal else 'error'
        print_to_screen(f'\t\t-> {msg}', level=lvl)
        if not equal:
            return False
        if not traj_skip:
            print_to_screen('\tComparing for trajectory archive:')
            for i in {'0_', '1_'}:
                archive_errors = soft_archive_comparison(
                    os.path.join(run1, ensemble, i + TRAJ),
                    os.path.join(run2, ensemble, i + TRAJ),
                )
                if archive_errors:
                    print_to_screen('\t\t-> Archives differ', level='error')
                    return False
            print_to_screen('\t\t-> Archives are equal', level='success')
        else:
            print_to_screen('\tSkipping comparison for trajectory archive.')
    return True


def main():
    """Parse arguments and execute the comparison.

    Returns
    -------
    out : int
        0 if the comparison was successful, 1 otherwise.

    """
    colorama.init(autoreset=True)
    parser = argparse.ArgumentParser()
    parser.add_argument('directories', metavar='N', type=str, nargs=2,
                        help='The directories to use for the comparison')
    parser.add_argument('--path_skip', nargs='+', type=int,
                        help='Columns to skip for pathensemble.txt')
    parser.add_argument('--energy_skip', nargs='+', type=str,
                        help='Columns to skip for energy.txt')
    parser.add_argument('--traj_skip', action='store_true',
                        help='Skip comparison for trajectory archives')
    args_dict = vars(parser.parse_args())

    run1 = args_dict['directories'][0]
    run2 = args_dict['directories'][1]

    settings = parse_settings_file(os.path.join(run1, RETIS_RST))
    inter = settings['simulation']['interfaces']
    for i, _ in enumerate(inter):
        ensemble_dir = generate_ensemble_name(i)
        result = compare_ensemble(
            run1, run2, ensemble_dir,
            path_skip=args_dict['path_skip'],
            energy_skip=args_dict['energy_skip'],
            traj_skip=args_dict['traj_skip'],
        )
        if not result:
            print_to_screen(
                f'Comparison failed for {ensemble_dir}. Aborting!',
                level='error',
            )
            return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module defining functions useful in the analysis of simulation data.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

running_average (:py:func:`.running_average`)
    Method to calculate a running average.

block_error (:py:func:`.block_error`)
    Perform block error analysis.

block_error_corr (:py:func:`.block_error_corr`)
    Method to run a block error analysis and calculate relative
    errors and correlation length.
"""
import numpy as np
from pyretis.analysis.histogram import histogram_and_avg

__all__ = ['running_average', 'block_error', 'block_error_corr',
           'analyse_data', 'mean_square_displacement']

np.seterr(divide='ignore', invalid='ignore')
np.set_printoptions(legacy='1.21')


def running_average(data):
    """Create a running average of the given data.

    The running average will be calculated over the rows.

    Parameters
    ----------
    data : numpy.array
        This is the data we will average.

    Returns
    -------
    out : numpy.array
        The running average.

    """
    if len(data.shape) == 2:
        w_data = data[:, 0]*data[:, 1]
        return w_data.cumsum() / data[:, 1].cumsum()
    return data.cumsum() / np.ones(data.shape[0]).cumsum()


def block_error(data, maxblock=None, blockskip=1, weights=None):
    """Perform block error analysis.

    This function will estimate the standard deviation in the input
    data by performing a block analysis. The number of blocks
    to consider can be specified or it will be taken as the
    half of the length of the input data. Averages and variance are
    calculated using an on-the-fly algorithm [1]_.

    Parameters
    ----------
    data : numpy.array (or iterable with data points)
        The data to analyse.
    maxblock : int, optional
        Can be used to set the maximum length of the blocks to
        consider. Note that the `maxblock` will never be set longer
        than half the length in data.
    blockskip : int, optional
        This can be used to skip certain block lengths, i.e.
        `blockskip = 1` will consider all blocks up to `maxblock`, while
        `blockskip = n` will consider every n'th block up to `maxblock`,
        i.e. it will use block lengths equal to `1`, `1 + n`, `1 + 2*n`,
        and so on.

    Returns
    -------
    blocklen : numpy.array
        These contain the block lengths considered.
    block_avg : numpy.array
        The averages as a function of the block length.
    block_err : numpy.array
        Estimate of errors as a function of the block length.
    block_err_avg : float
        Average of the error estimate using blocks where
        ``length > maxblock//2``.

    References
    ----------
    .. [1] Wikipedia, "Algorithms for calculating variance",
       http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance

    """
    if maxblock is None or maxblock < 1:
        maxblock = len(data) // 2
    else:
        maxblock = min(maxblock, len(data) // 2)
    indata = data
    if weights is None:
        if len(data.shape) == 1:
            weights = np.ones(len(data))*2
        else:
            data = (i[0] for i in indata)
            weights = (i[1] for i in indata)
    # define helper variables:
    blocklen = np.arange(0, maxblock, blockskip, dtype='i4')
    # blocklen contains the lengths of the blocks
    blocklen += 1
    # +1 to make blocklen[i] = length of block no i where numbering
    # starts at 0 -> blocklen[0] = 1 and so on. Note that arange does
    # create [0, ..., maxblock).
    block = np.zeros(len(blocklen))  # to accumulate values for a block
    blockw = np.zeros(len(blocklen))  # accumulate weights for a block
    tot_w = np.zeros(block.shape)  # accumulate total weights
    tot_w_s = np.zeros(block.shape)  # accumulate total weights**2
    block_avg = np.zeros(block.shape)  # to store averages in block
    block_var = np.zeros(block.shape)  # estimator of variance

    # Algorithm taken from
    # https://markusthill.github.io/math/stats/ml/
    # online-estimation-of-weighted-sample-mean-and-coviarance-matrix/
    # equation 39 where we delay the normalization to the end
    for i, (datai, weighti) in enumerate(zip(data, weights)):
        block += weighti*datai  # accumulate the value to all blocks
        blockw += weighti
        # next pick out blocks which are "full":
        k = np.where((i + 1) % blocklen == 0)[0]
        # update estimate of average and variance
        block[k] = np.nan_to_num(block[k] / blockw[k])  # catch 0/0 = NaN
        tot_w[k] += blockw[k]
        tot_w_s[k] += blockw[k]**2
        delta1 = blockw[k] * (block[k] - block_avg[k])
        block_avg[k] = block_avg[k] + np.nan_to_num(delta1 / tot_w[k])
        block_var[k] = block_var[k] + delta1 * (block[k] - block_avg[k])
        # reset these blocks
        block[k] = 0.0
        blockw[k] = 0

    nsamp = tot_w * tot_w / tot_w_s
    block_var *= nsamp / (nsamp - 1) / tot_w
    block_err = np.sqrt(block_var / nsamp)  # estimate of error
    block_err_avg = np.average(
        block_err[np.where(blocklen > maxblock // 2)[0]])
    return blocklen, block_avg, block_err, block_err_avg


def block_error_corr(data, maxblock=None, blockskip=1):
    """Run block error analysis on the given data.

    This will run the block error analysis and return the relative
    errors and correlation length.

    Parameters
    ----------
    data : numpy.array
        Data to analyse.
    maxblock : int, optional
        The maximum block length to consider.
    blockskip : int, optional
        This can be used to skip certain block lengths, i.e.
        `blockskip = 1` will consider all blocks up to `maxblock`, while
        `blockskip = n` will consider every n'th block up to `maxblock`,
        i.e. it will use block lengths equal to `1`, `1 + n`, `1 + 2*n`,
        and so on.

    Returns
    -------
    out[0] : numpy.array
        These contains the block lengths considered (`blen`).
    out[1] : numpy.array
        Estimate of errors as a function of the block length (`berr`).
    out[2] : float
        Average of the error estimate for blocks (`berr_avg`)
        with ``length > maxblock // 2``.
    out[3] : numpy.array
        Estimate of relative errors normalised by the overall average
        as a function of block length (`rel_err`).
    out[4] : float
        The average relative error (`avg_rel_err`), for blocks
        with ``length > maxblock // 2``.
    out[5] : numpy.array
        The estimated correlation length as a function of the block
        length (`ncor`).
    out[6] : float
        The average (for blocks with length > maxblock // 2) estimated
        correlation length (`avg_ncor`).

    """
    blen, bavg, berr, berr_avg = block_error(data, maxblock=maxblock,
                                             blockskip=blockskip)
    # also calculate some relative errors:
    rel_err = np.divide(berr, abs(bavg[0]))
    avg_rel_err = np.divide(berr_avg, abs(bavg[0]))
    ncor = np.divide(berr**2, berr[0]**2)
    avg_ncor = np.divide(berr_avg**2, berr[0]**2)
    return blen, berr, berr_avg, rel_err, avg_rel_err, ncor, avg_ncor


def mean_square_displacement(data, ndt=None):
    """Calculate the mean square displacement for the given data.

    Parameters
    ----------
    data : numpy.array, 1D
        This numpy.array contains the data as a function of time.
    ndt : int, optional
        This parameter is the number of time origins. I.e. points up to
        ndt will be used as time origins. If not specified the value of
        the input ``data.size // 5`` will be used.

    Returns
    -------
    msd : numpy.array, 2D
        The first column is the mean squared displacement and the
        second column is the corresponding standard deviation.

    """
    length = data.shape[0]
    if ndt is None or ndt < 1:
        ndt = length // 5
    msd = []
    for i in range(1, ndt):
        delta = (data[i:] - data[:-i])**2
        msd.append((delta.mean(), delta.std()))
    return np.array(msd)


def analyse_data(data, settings):
    """Analyse the given data and run some common analysis procedures.

    Specifically, it will:

    1) Calculate a running average.

    2) Obtain a histogram.

    3) Run a block error analysis.


    Parameters
    ----------
    data : numpy.array, 1D
        This numpy.array contains the data as a function of time.
    settings : dict
        This dictionary contains settings for the analysis.

    Returns
    -------
    result : dict
        This dict contains the results.

    """
    result = {}
    asett = settings['analysis']
    # 1) Do the running average
    result['running'] = running_average(data)
    # 2) Obtain distributions:
    result['distribution'] = histogram_and_avg(data, asett['bins'],
                                               density=True)
    # 3) Do the block error analysis:
    result['blockerror'] = block_error_corr(data,
                                            maxblock=asett['maxblock'],
                                            blockskip=asett['blockskip'])
    return result



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of a class for handling output related to simulations.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Task (:py:class:`.Task`)
    Base class for tasks. This is used by :py:class:`.SimulationTask`
    and :py:class:`.OutputTask`.

OutputTask (:py:class:`.OutputTask`)
    A class representing a simulation output task.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

get_task_type (:py:meth:`.get_task_type`)
    Do additional handling for a path task.

get_file_mode (:py:meth:`.get_file_mode`)
    Determine if we should append or backup existing files.

task_from_settings (:py:meth:`.task_from_settings`)
    Create output task from simulation settings.

Important variables defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OUTPUT_TASKS (:py:data:`.OUTPUT_TASKS`)
    A dictionary defining the different output tasks known to PyRETIS.

"""
import logging
from pyretis.core.common import initiate_instance
from pyretis.inout.common import generate_file_name
from pyretis.inout.formats import (
    EnergyFormatter,
    OrderFormatter,
    CrossFormatter,
    SnapshotFormatter,
    ThermoTableFormatter,
    PathTableFormatter,
    EnergyPathFormatter,
    OrderPathFormatter,
    PathIntFormatter,
    PathEnsembleFormatter,
    RETISResultFormatter,
)
from pyretis.inout.fileio import FileIO
from pyretis.inout.screen import ScreenOutput
from pyretis.inout.archive import PathStorage
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


OUTPUT_TASKS = {}
"""Define a set of known output tasks.

The output tasks are defined as dictionaries with the following keys:

* target : string
    "file" or "screen", defines where the task writes to.
* filename : string
    A default file name for an output file if writing to a file.
* result : tuple of strings
    Determines what item from the result dictionary we are outputting.
* when : string
    Determines what input setting from the "output" section is used to
    define the output frequency. Default values are defined by the
    output section, see: py:mod:`pyretis.inout.settings.settings`.
* formatter : object like :py:class:`.OutputFormatter`
    Selects the formatter for the output.
* writer : object like :py:class:`.OutputBase`
    Selects the writer for the output.
* settings : tuple of strings, optional
    A dict with additional settings which can be passed to the
    formatter if needed. These settings can, for instance, be
    defined in the output section of the input file.

The writer can be defined explicitly, or via the formatter. If a
formatter is given, then the generic :py:class:`.FileIO` will be used.
If no formatter is given, then the writer is assumed to be given.

"""
OUTPUT_TASKS['energy'] = {
    'target': 'file',
    'filename': 'energy.txt',
    'result': ('thermo',),
    'when': 'energy-file',
    'formatter': EnergyFormatter,
}
OUTPUT_TASKS['order'] = {
    'target': 'file',
    'filename': 'order.txt',
    'result': ('order',),
    'when': 'order-file',
    'formatter': OrderFormatter,
}
OUTPUT_TASKS['cross'] = {
    'target': 'file',
    'filename': 'cross.txt',
    'result': ('cross',),
    'when': 'cross-file',
    'formatter': CrossFormatter,
}
OUTPUT_TASKS['traj-txt'] = {
    'target': 'file',
    'filename': 'traj.txt',
    'result': ('system',),
    'when': 'trajectory-file',
    'formatter': SnapshotFormatter,
}
OUTPUT_TASKS['traj-xyz'] = {
    'target': 'file',
    'filename': 'traj.xyz',
    'result': ('system',),
    'when': 'trajectory-file',
    'formatter': SnapshotFormatter,
}
OUTPUT_TASKS['thermo-screen'] = {
    'target': 'screen',
    'result': ('thermo',),
    'when': 'screen',
    'formatter': ThermoTableFormatter,
}
OUTPUT_TASKS['thermo-file'] = {
    'target': 'file',
    'filename': 'thermo.txt',
    'result': ('thermo',),
    'when': 'energy-file',
    'formatter': ThermoTableFormatter,
}
OUTPUT_TASKS['pathensemble'] = {
    'target': 'file',
    'filename': 'pathensemble.txt',
    'result': ('pathensemble',),
    'when': 'pathensemble-file',
    'formatter': PathEnsembleFormatter,
}
OUTPUT_TASKS['pathensemble-screen'] = {
    'target': 'screen',
    'result': ('pathensemble',),
    'when': 'screen',
    'formatter': PathTableFormatter,
}
OUTPUT_TASKS['pathensemble-retis-screen'] = {
    'target': 'screen',
    'result': ('pathensemble',),
    'when': 'screen',
    'formatter': RETISResultFormatter,
}
OUTPUT_TASKS['path-order'] = {
    'target': 'file',
    'filename': 'order.txt',
    'result': ('path', 'status'),
    'when': 'order-file',
    'formatter': OrderPathFormatter,
}
OUTPUT_TASKS['path-energy'] = {
    'target': 'file',
    'filename': 'energy.txt',
    'result': ('path', 'status'),
    'when': 'energy-file',
    'formatter': EnergyPathFormatter,
}
OUTPUT_TASKS['path-traj-int'] = {
    'target': 'file',
    'filename': 'traj.txt',
    'result': ('path', 'status'),
    'when': 'trajectory-file',
    'formatter': PathIntFormatter,
}
OUTPUT_TASKS['path-traj-ext'] = {
    'target': 'file-archive',
    'filename': 'traj.txt',
    'result': ('path', 'status', 'pathensemble'),
    'when': 'trajectory-file',
    'writer': PathStorage,
}


class Task:
    """Base representation of a "task".

    A task is just something that is supposed to be executed at
    a certain point. This class will just set up functionality
    that is common for output tasks and for simulation tasks.

    Attributes
    ----------
    when : dict
        Determines when the task should be executed.

    """

    _ALLOWED_WHEN = {'every', 'at'}

    def __init__(self, when):
        """Initialise the task.

        Parameters
        ----------
        when : dict, optional
            Determines if the task should be executed.

        """
        self._when = None
        self.when = when

    @property
    def when(self):
        """Return the "when" property."""
        return self._when

    @when.setter
    def when(self, when):
        """Update self to new value(s).

        It will only update `self.when` for the keys given in the
        input `when`.

        Parameters
        ----------
        when : dict
            This dict contains the settings to update.

        Returns
        -------
        out : None
            Returns `None` but modifies `self.when`.

        """
        if when is None:
            self._when = when
        else:
            if self._when is None:
                self._when = {}
            for key, val in when.items():
                if key in self._ALLOWED_WHEN:
                    self._when[key] = val
                else:
                    logger.warning(
                        'Ignoring unknown "when" setting: "%s"', key
                    )

    def execute_now(self, step):
        """Determine if a task should be executed.

        Parameters
        ----------
        step : dict of ints
            Keys are 'step' (current cycle number), 'start' cycle number at
            start 'stepno' the number of cycles we have performed so far.

        Returns
        -------
        out : boolean
            True of the task should be executed.

        """
        if self.when is None:
            return True
        exe = False
        if 'every' in self.when:
            exe = step['stepno'] % self.when['every'] == 0
        if not exe and 'at' in self.when:
            try:
                exe = step['step'] in self.when['at']
            except TypeError:
                exe = step['step'] == self.when['at']
        return exe

    def task_dict(self):
        """Return basic info about the task."""
        return {'when': self.when}


class OutputTask(Task):
    """A base class for simulation output.

    This class will handle an output task for a simulation. The
    output task consists of one object which is responsible for
    formatting the output data and one object which is responsible
    for writing that data, for instance to the screen or to a file.

    Attributes
    ----------
    target : string
        This string identifies what kind of output we are dealing with.
        This will typically be either "screen" or "file".
    name : string
        This string identifies the task, it can, for instance, be used
        to reference the dictionary used to create the writer.
    result : tuple of strings
        This string defines the result we are going to output.
    writer : object like :py:class:`.OutputBase`
        This object will handle the actual outputting
        of the result.
    when : dict
        Determines if the task should be executed.

    """

    def __init__(self, name, result, writer, when):
        """Initialise the generic output task.

        Parameters
        ----------
        name : string
            This string identifies the task, it can, for instance, be used
            to reference the dictionary used to create the writer.
        result : list of strings
            These strings define the results we are going to output.
        writer : object like :py:class:`.IOBase`
            This object will handle formatting of the actual result and
            output to screen or to a file.
        when : dict
            Determines when the output should be written. Example:
            `{'every': 10}` will be executed at every 10th step.

        """
        super().__init__(when)
        self.target = writer.target
        self.name = name
        self.result = result
        self.writer = writer

    def output(self, simulation_result):
        """Output given results from simulation steps.

        This will output the task using the result found in the
        `simulation_result` which should be the dictionary returned
        from a simulation object (e.g. object like
        :py:class:`.Simulation`) after a step.

        Parameters
        ----------
        simulation_result : dict
            This is the result from a simulation step.

        Returns
        -------
        out : boolean
            True if the writer wrote something, False otherwise.

        """
        step = simulation_result['cycle']
        if not self.execute_now(step):
            return False
        result = []
        for res in self.result:
            if res not in simulation_result:
                return False  # Requested result was not ready at this step.
            result.append(simulation_result[res])
        if len(result) == 1:
            return self.writer.output(step['step'], result[0])
        return self.writer.output(step['step'], result)

    def __str__(self):
        """Print information about the output task."""
        msg = [f'Output task {self.name}']
        msg += [f'- Output frequency: {self.when}']
        msg += [f'- Acting on result(s): {self.result}']
        msg += [f'- Target is: {self.target}']
        msg += [f'- Writer is:\xa0{self.writer}']
        return '\n'.join(msg)

    def task_dict(self):
        """Return a dict with info about the task."""
        return {'name': self.name,
                'when': self.when,
                'result': self.result,
                'target': self.target,
                'writer': self.writer.__class__,
                'formatter': self.writer.formatter_info()}


def get_task_type(task, engine):
    """Do additional handling for a path task.

    The path task is special since we do very different things for
    external paths. The set-up required to do this is handled here.

    Parameters
    ----------
    task : dict
        Settings related to the specific task.
    engine : object like :py:class:`.EngineBase`
        This object is used to determine if we need to do something
        special for external engines. If no engine is given, we do
        not do anything special.

    Returns
    -------
    out : string
        The task type we are going to be creating for.

    """
    if task['type'] == 'path-traj-{}':
        if engine is None or engine.engine_type in ('internal', 'openmm'):
            fmt = 'int'
        else:
            fmt = 'ext'
        return task['type'].format(fmt)
    return task['type']


def get_file_mode(settings):
    """Determine if we should append or backup existing files.

    This method translates the backup settings into a file mode string.
    We assume here that the file is opened for writing.

    Parameters
    ----------
    settings : dict
        The simulation settings.

    Returns
    -------
    file_mode : string
        A string representing the file mode to use.

    """
    file_mode = 'w'
    try:
        old = settings['output']['backup'].lower()
        if old == 'append':
            logger.debug('Will append to existing files.')
            file_mode = 'a'
    except AttributeError:
        logger.warning('Could not understand setting for "backup"'
                       ' in "output" section.')
        old = 'backup'
        logger.warning('Handling of existing files is set to: "%s"', old)
        settings['output']['backup'] = old
    return file_mode


def task_from_settings(task, settings, directory, engine, progress=False):
    """Create output task from simulation settings.

    Parameters
    ----------
    task : dict
        Settings for creating a task. This dict contains the type and
        name of the task to create. It can also contain overrides to
        the default settings in :py:data:`OUTPUT_TASKS`.
    settings : dict
        Settings for the simulation.
    directory : string
        The directory to write output files to.
    engine : object like :py:class:`.EngineBase`
        This object is used to determine if we need to do something
        special for external engines. If no engine is given, we do
        not do anything special.
    progress : boolean, optional
        For some simulations, the user may select to display a
        progress bar. We will then just disable the other screen
        output.

    Returns
    -------
    out : object like :py:class:`.OutputTask`
        An output task we can use in the simulation.

    """
    task_type = get_task_type(task, engine)
    task_settings = OUTPUT_TASKS[task_type].copy()
    # Override defaults if any:
    for key in task_settings:
        if key in task:
            task_settings[key] = task[key]

    when = {'every': settings['output'][task_settings['when']]}
    if when['every'] < 1:
        logger.info('Skipping output task %s (freq < 1)', task_type)
        return None

    target = task_settings['target']
    if target == 'screen' and progress:
        logger.info(
            'Disabling output to screen %s since progress bar is ON',
            task['name'],
        )
        return None
    formatter = None
    # Initiate the formatter, note that here we can customize the formatter
    # by supplying arguments to it. This was supported in a previous version
    # of PyRETIS, but for now, none of the formatters needs settings to be
    # created.
    if task_settings.get('formatter', None) is not None:
        formatter = initiate_instance(
            task_settings['formatter'],
            task_settings.get('formatter-settings', {}),
        )
    # Create writer:
    writer = None
    if target == 'screen':
        klass = task_settings.get('writer', ScreenOutput)
        writer = klass(formatter)
    if target in ('file', 'file-archive'):
        filename = generate_file_name(task_settings['filename'], directory,
                                      settings)
        file_mode = get_file_mode(settings)
        if target == 'file':
            klass = task_settings.get('writer', FileIO)
            writer = klass(filename, file_mode, formatter, backup=True)
        if target == 'file-archive':
            klass = task_settings.get('writer', PathStorage)
            writer = klass()
    # Finally make the output task:
    if writer is not None:
        return OutputTask(task['name'], task_settings['result'], writer, when)
    logger.warning('Unknown target "%s". Ignoring task: %s',
                   target, task_type)
    return None



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This is an example of running an Umbrella Window simulation."""
import numpy as np
from matplotlib import pyplot as plt
from pyretis.core import System, RandomGenerator, Particles
from pyretis.simulation.mc_simulation import UmbrellaWindowSimulation
from pyretis.forcefield import ForceField
from pyretis.forcefield.potentials import DoubleWell, RectangularWell
from pyretis.analysis import histogram, match_all_histograms


mysystem = System(temperature=500, units='eV/K')
mysystem.particles = Particles(dim=1)
mysystem.add_particle(name='X', pos=np.array([-0.7]))

potential_dw = DoubleWell(a=1, b=1, c=0.02)
potential_rw = RectangularWell()
forcefield = ForceField(desc='Double well', potential=[potential_dw])
forcefield_bias = ForceField(desc='Double well with rectangular bias',
                             potential=[potential_dw, potential_rw])
mysystem.forcefield = forcefield_bias

umbrellas = [[-1.0, -0.4], [-0.5, -0.2], [-0.3, 0.0],
             [-0.1, 0.2], [0.1, 0.4], [0.3, 0.6], [0.5, 1.0]]
n_umb = len(umbrellas)
MINCYCLES = 1e4  # Number of MC steps to perform.
MAXDX = 0.1  # Maximum allowed displacement in the MC step(s).

RANDSEED = 1  # Seed for the random number generator:
RGEN = RandomGenerator(seed=RANDSEED)

trajectory, energy = [], []  # For storing trajectories & energies.
for i, umbrella in enumerate(umbrellas):
    print('Running umbrealla no: {} of {}. Location: {}'.format(i + 1, n_umb,
                                                                umbrella))
    # Move rectangular potential to correct place:
    params = {'left': umbrella[0], 'right': umbrella[1]}
    mysystem.forcefield.update_potential_parameters(potential_rw, params)
    mysystem.potential()  # Re-calculate potential energy.
    over = umbrellas[min(i + 1, n_umb - 1)][0]  # Position we must cross.
    simulation = UmbrellaWindowSimulation(
        ensemble={
            'system': mysystem,
            'rgen': RGEN
        },
        settings={
            'simulation': {
                'task': 'umbrella',
                'umbrella': umbrella,
                'overlap': over,
                'maxdx': MAXDX
            }
        },
        controls={'steps': MINCYCLES}
    )
    # Also create empy list for storing some data:
    traj, ener = [], []
    for result in simulation.run():
        for pos in mysystem.particles.pos:
            traj.append(pos)
            ener.append(mysystem.particles.vpot)
    trajectory.append(np.array(traj))
    energy.append(np.array(ener))
    print('Done. Cycles: {}'.format(simulation.cycle['stepno']))

# We can now post-process the simulation output:
BINS = 100
LIM = (-1.1, 1.1)
histograms = [histogram(traj, bins=BINS, limits=LIM) for traj in trajectory]
# Extract the bins (the midpoints) and the bin-width:
bin_x = histograms[0][-1]
dbin = bin_x[1] - bin_x[0]
# We are going to match these histograms:
print('Matching histograms...')
histograms_s, _, hist_avg = match_all_histograms(histograms, umbrellas)

print('Plotting matched histograms')
fig = plt.figure()
axs = fig.add_subplot(111)
axs.set_yscale('log')
axs.set_xlabel('Position ($x$)', fontsize='large')
axs.set_ylabel('Matched histograms', fontsize='large')
colors = ['blue', 'green', 'darkviolet', 'brown', 'gray', 'crimson', 'cyan']
for i, histo in enumerate(histograms_s):
    axs.bar(bin_x - 0.5 * dbin, histo, dbin, color=colors[i],
            alpha=0.6, log=True)
axs.plot(bin_x, hist_avg, lw=7, color='orangered', alpha=0.6,
         label='Average after matching')
axs.legend()
plt.xlim((-1.1, 1.1))

print('Plotting the free energy')
fig2 = plt.figure()
ax2 = fig2.add_subplot(111)
XPOT = np.linspace(-2, 2, 1000)
free = -np.log(hist_avg) / mysystem.temperature['beta']  # The free energy.
# Set up unbiased potential for plotting:
VPOT = []
for xi in XPOT:
    mysystem.particles.pos = xi
    VPOT.append(forcefield.evaluate_potential(mysystem))
VPOT = np.array(VPOT)
free += (VPOT.min() - free.min())
ax2.plot(XPOT, VPOT, 'blue', lw=3, label='Unbiased potential', alpha=0.5)
ax2.plot(bin_x, free, lw=7, alpha=0.5, color='green', label='Free energy')
ax2.set_xlabel('Position ($x$)', fontsize='large')
ax2.set_ylabel('Potential energy ($V(x)$) / eV', fontsize='large')
ax2.legend()
plt.xlim((-1.1, 1.1))
plt.ylim((-0.3, 0.05))
plt.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains common functions for the input/output.

It contains some slave functions that are used in the in/output function
of PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OutputBase (:py:class:`.OutputBase`)
    A base class for handling the output.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

check_python_version (:py:func:`.check_python_version`)
    A method that will give warnings when we use older and possibly
    unsupported Python versions.

create_backup (:py:func:`.create_backup`)
    A function to handle the creation of backups of old files.

make_dirs (:py:func:`.make_dirs`)
    Create directories (for path simulation).

create_empty_ensembles (:py:func:`.create_ensembles`)
    A method to prepare the ensembles inputs in settings

simplify_ensemble_name (:py:func:`.simplify_ensemble_name`)
    Simplify the name of ensembles for creating directories.

generate_file_name (:py:func:`.generate_file_name`)
    Generate file name for an output task, from settings.

"""
import logging
import errno
import os
import re
import sys
from abc import ABCMeta, abstractmethod
from pyretis import MIN_PYTHON_VERSION
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = [
    'check_python_version',
    'create_backup',
    'create_empty_ensembles',
    'make_dirs',
    'OutputBase',
    'simplify_ensemble_name',
    'add_dirname',
    'name_file',
    'generate_file_name',
    'TRJ_FORMATS'
]


# Hard-coded patters for energy analysis output files.
# These are just used to make it simpler to change these default
# names in the future.
ENERFILES = {'energies': 'energies',
             'run_energies': 'runenergies',
             'temperature': 'temperature',
             'run_temp': 'runtemperature',
             'block': '{}block',
             'dist': '{}dist'}
# hard-coded information for the energy terms:
ENERTITLE = {'vpot': 'Potential energy',
             'ekin': 'Kinetic energy',
             'etot': 'Total energy',
             'ham': 'Hamilt. energy',
             'temp': 'Temperature',
             'elec': 'Energy (externally computed)'}
# hard-coded patters for flux analysis output files:
FLUXFILES = {'runflux': 'runflux_{}',
             'block': 'errflux_{}'}
# order files:
ORDERFILES = {'order': 'order',
              'ordervel': 'orderv',
              'run_order': 'runorder',
              'dist': 'orderdist',
              'block': 'ordererror',
              'msd': 'ordermsd'}
# hard-coded patters for path analysis output files:
PATHFILES = {'pcross': '{}_pcross',
             'prun': '{}_prun',
             'prun_sl': '{}_prun_sl',
             'prun_sr': '{}_prun_sr',
             'perror': '{}_perror',
             'perror_sl': '{}_perror_sl',
             'perror_sr': '{}_perror_sr',
             'lengtherror': '{}_lerror',
             'pathlength': '{}_lpath',
             'shoots': '{}_shoots'}
# hard-coded patterns for matched files:
PATH_MATCH = {'total': 'total-probability',
              'progress': 'overall-rrun',
              'error': 'overall-err',
              'match': 'matched-probability'}
# Supported trj formats # TODO (LAMMPS missing)
TRJ_FORMATS = ['.xyz', '.gro', '.trr', '.txt', '.g96']


def create_backup(outputfile):
    """Check if a file exist and create backup if requested.

    This function will check if the given file name exists and if it
    does, it will move that file to a new file name such that the given
    one can be used without overwriting.

    Parameters
    ----------
    outputfile : string
        This is the name of the file we wish to create.

    Returns
    -------
    out : string
        This string is None if no backup is made, otherwise, it will
        just say what file was moved (and to where).

    Note
    ----
    No warning is issued here. This is just in case the `msg` returned
    here will be part of some more elaborate message.

    """
    filename = f'{outputfile}'
    fileid = 0
    msg = None
    while os.path.isfile(filename) or os.path.isdir(filename):
        filename = f'{outputfile}_{fileid:03d}'
        fileid += 1
    if fileid > 0:
        msg = f'Backup existing file "{outputfile}" to "{filename}"'
        os.rename(outputfile, filename)
    return msg


def make_dirs(dirname):
    """Create directories for path simulations.

    This function will create a folder using a specified path.
    If the path already exists and if it's a directory, we will do
    nothing. If the path exists and is a file we will raise an
    `OSError` exception here.

    Parameters
    ----------
    dirname : string
        This is the directory to create.

    Returns
    -------
    out : string
        A string with some info on what this function did. Intended for
        output.

    """
    try:
        os.makedirs(dirname)
        msg = f'Created directory: "{dirname}"'
    except OSError as err:
        if err.errno != errno.EEXIST:  # pragma: no cover
            raise err
        if os.path.isfile(dirname):
            msg = f'"{dirname}" is a file. Will abort!'
            raise OSError(errno.EEXIST, msg) from err
        if os.path.isdir(dirname):
            msg = f'Directory "{dirname}" already exist.'
    return msg


def simplify_ensemble_name(ensemble, fmt='{:03d}'):
    """Simplify path names for file/directory names.

    Here, we are basically translating ensemble names to more friendly
    names for directories and files that is:

    - ``[0^-]`` returns ``000``,
    - ``[0^+]`` returns ``001``,
    - ``[1^+]`` returns ``002``, etc.

    Parameters
    ----------
    ensemble : string
        This is the string to simplify.
    fmt : string. optional
        This is a format to use for the directories.

    """
    match_ensemble = re.search(r'(?<=\[)(\d+)(?=\^)', ensemble)
    if match_ensemble:
        ens = int(match_ensemble.group())
    else:
        match_ensemble = re.search(r'(?<=\[)(\d+)(?=\])', ensemble)
        if match_ensemble:
            ens = int(match_ensemble.group())
        else:
            return ensemble  # Assume that the ensemble is OK as it is.
    match_dir = re.search(r'(?<=\^)(.)(?=\])', ensemble)
    if match_dir:
        dire = match_dir.group()
        if dire != '-':
            ens += 1
    else:
        msg = [f'Could not get direction for ensemble {ensemble}.',
               'We assume "+", note that this might overwrite files']
        logger.warning('\n'.join(msg))
        ens += 1
    return fmt.format(ens)


def add_dirname(filename, dirname):
    """Add a directory as a prefix to a filename, i.e. `dirname/filename`.

    Parameters
    ----------
    filename : string
        The filename.
    dirname : string
        The directory we want to prefix. It can be None, in which
        case we ignore it.

    Returns
    -------
    out : string
        The path to the resulting file.

    """
    if dirname is not None:
        return os.path.join(dirname, filename)
    return filename


def name_file(name, extension, path=None):
    """Return a file name by joining a name and an file extension.

    This function is used to create file names. It will use
    `os.extsep` to create the file names and `os.path.join` to add a
    path name if the `path` is given. The returned file name will be of
    form (example for posix): ``path/name.extension``.

    Parameters
    ----------
    name : string
        This is the name, without extension, for the file.
    extension : string
        The extension to use for the file name.
    path : string, optional
        An optional path to add to the file name.

    Returns
    -------
    out : string
        The resulting file name.

    """
    return add_dirname(os.extsep.join([name, extension]), path)


def generate_file_name(basename, directory, settings):
    """Generate file name for an output task, from settings.

    Parameters
    ----------
    basename : string
        The base file name to use.
    directory : string
        A directory to output to. Can be None to output to the
        current working directory.
    settings : dict
        The input settings

    Returns
    -------
    filename : string
        The file name to use.

    """
    prefix = settings['output'].get('prefix', None)
    if prefix is not None:
        filename = f'{prefix}{basename}'
    else:
        filename = basename
    filename = add_dirname(filename, directory)
    return filename


def check_python_version():
    """Give a warning about old python version(s)."""
    version = sys.version_info[:2]  # Only check major, minor
    if version < MIN_PYTHON_VERSION:
        msgtxt = ('Please upgrade to Python {}.{}.'
                  '\nPython {}.{} is not supported!')
        msgtxt = msgtxt.format(*MIN_PYTHON_VERSION,
                               *version)
        logger.error(msgtxt)
        raise SystemExit(msgtxt)


class OutputBase(metaclass=ABCMeta):
    """A generic class for handling output.

    Attributes
    ----------
    formatter : object like py:class:`.OutputFormatter`
        The object responsible for formatting output.
    target : string
        Determines where the target for the output, for
        instance "screen" or "file".
    first_write : boolean
        Determines if we have written something yet, or
        if this is the first write.

    """

    target = None

    def __init__(self, formatter):
        """Create the object and attach a formatter."""
        self.formatter = formatter
        self.first_write = True

    def output(self, step, data):
        """Use the formatter to write data to the file.

        Parameters
        ----------
        step : int
            The current step number.
        data : list
            The data we are going to output.

        """
        if self.first_write and self.formatter.print_header:
            self.first_write = False
            self.write(self.formatter.header)
        for line in self.formatter.format(step, data):
            self.write(line)

    @abstractmethod
    def write(self, towrite, end='\n'):
        """Write a string to the output defined by this class.

        Parameters
        ----------
        towrite : string
            The string to write.
        end : string, optional
            A "terminator" for the given string.

        Returns
        -------
        status : boolean
            True if we managed to write, False otherwise.

        """
        return

    def formatter_info(self):
        """Return a string with info about the formatter."""
        if self.formatter is not None:
            return self.formatter.__class__
        return None

    def __str__(self):
        """Return basic info."""
        return f'{self.__class__.__name__}\n\t* Formatter: {self.formatter}'


def create_empty_ensembles(settings):
    """Create missing ensembles in the settings.

    Checks the input and allocate it to the right ensemble. In theory
    inouts shall include all these info, but it is not practical.

    Parameters
    ----------
    settings : dict
        The current input settings.

    Returns
    -------
    None, but this method might add data to the input settings.

    """
    ints = settings['simulation']['interfaces']

    # Determine how many ensembles are needed.
    add0 = 0
    if not settings['simulation'].get('flux', False):
        add0 += 1
    if not settings['simulation'].get('zero_ensemble', True):
        add0 += 1

    # if some ensembles have inputs, they need to be kept.
    if 'ensemble' in settings:
        orig_set = settings['ensemble'].copy()
    else:
        orig_set = []

    settings['ensemble'] = []
    add = add0

    # if in the main settings an ensemble_number is defined, then only
    # that ensemble will be considered.
    if 'tis' in settings:
        idx = settings['tis'].get('ensemble_number')
        detect = settings['tis'].get('detect', ints[2])
        if idx is not None:
            settings['ensemble'].append({'interface': ints[1],
                                         'tis': {'ensemble_number': idx,
                                                 'detect': detect}})
            for sav in orig_set:
                settings['ensemble'][0] = {**settings['ensemble'][0], **sav}
            return
    # if one wants to compute the flux, the 000 ensemble is for it
    # todo remove this labelling mismatch, and give to the flux
    # a flux name folder (instead of 000), and leave 000 for the O^+ ens.
    if settings['simulation'].get('flux', False):
        settings['ensemble'].append({'interface': ints[0],
                                     'tis': {'ensemble_number': 0,
                                             'detect': ints[1]}})
        add = add0 + 1
    for i in range(add, len(ints)):
        settings['ensemble'].append({'interface': ints[i - 1],
                                     'tis': {'ensemble_number': i,
                                             'detect': ints[i]}})

    # create the ensembles in setting, keeping eventual inputs.
    # nb. in the settings, specific input for an ensemble can be now given.
    for i_ens, ens in enumerate(settings['ensemble']):
        for sav in orig_set:
            if 'tis' in sav and 'ensemble_number' in sav['tis']:
                if ens['tis']['ensemble_number'] ==\
                        sav['tis']['ensemble_number']:
                    settings['ensemble'][i_ens].update(sav)
            elif ens['interface'] == sav['interface']:
                settings['ensemble'][i_ens].update(sav)

    return



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This package defines common methods which are used for testing.

Package structure
-----------------

Modules
~~~~~~~

compare.py (:py:mod:`pyretis.testing.compare`)
    Common methods for comparing results.

helpers.py (:py:mod:`pyretis.testing.helpers`)
    Common methods for tests.

"""



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Some common methods for the engine tests."""
import os
from pyretis.core.system import System
from pyretis.core.particles import ParticlesExt
from pyretis.engines.openmm import HAS_OPENMM
if HAS_OPENMM:
    try:
        import openmm as mm
        from openmm import app
        from openmm import unit
        from simtk import unit
    except ImportError:  # Openmm < 7.6
        from simtk.openmm import app
        import simtk.openmm as mm
        from simtk import unit
else:
    mm, app, unit = None, None, None

PDB_STRING = (
    """CRYST1   68.478   68.478   68.472  60.00  60.00  90.00 P 1           1
MODEL        0
ATOM      1  O   HOH A   1      38.510  67.082  29.996  1.00  0.00           O
ATOM      2  H1  HOH A   1      39.260  66.728  29.518  1.00  0.00           H
ATOM      3  H2  HOH A   1      38.877  67.385  30.827  1.00  0.00           H
ATOM      4  O   HOH A   2      19.361  51.482   4.207  1.00  0.00           O
ATOM      5  H1  HOH A   2      19.004  51.112   5.014  1.00  0.00           H
ATOM      6  H2  HOH A   2      20.299  51.294   4.252  1.00  0.00           H
ENDMDL
END
""")


def remove_dir(dirname):
    """Remove a directory silently."""
    try:
        os.removedirs(dirname)
    except OSError:
        pass


def make_test_system(conf):
    """Make a system with particles for testing."""
    system = System()
    system.particles = ParticlesExt(dim=3)
    system.particles.set_pos(conf)
    system.particles.set_vel(False)
    return system


def create_openmm_simulation(pdb='test_openmm_pyretis.pdb'):
    """
    This creates a default OpenMM simulation object for testing purposses.
    Adapted from builder.openmm.org
    """
    pdb = app.PDBFile(pdb)
    forcefield = app.ForceField('amber99sbildn.xml', 'tip3p.xml')

    system = forcefield.createSystem(pdb.topology, nonbondedMethod=app.PME,
                                     nonbondedCutoff=1.0*unit.nanometers,
                                     constraints=app.HBonds, rigidWater=True,
                                     ewaldErrorTolerance=0.0005)
    integrator = mm.LangevinIntegrator(300*unit.kelvin,
                                       1.0/unit.picoseconds,
                                       2.0*unit.femtoseconds)
    integrator.setConstraintTolerance(0.00001)

    simulation = app.Simulation(pdb.topology, system, integrator)
    simulation.context.setPositions(pdb.positions)
    return simulation


def write_openmm_simulation(name, pdb='test_openmm_pyretis.pdb'):
    CODE = (
        f"""
from openmm import app
import openmm as mm
from openmm import unit

pdb = app.PDBFile("{pdb}")
forcefield = app.ForceField('amber99sbildn.xml', 'tip3p.xml')

system = forcefield.createSystem(pdb.topology, nonbondedMethod=app.PME,
                                 nonbondedCutoff=1.0*unit.nanometers,
                                 constraints=app.HBonds, rigidWater=True,
                                 ewaldErrorTolerance=0.0005)
integrator = mm.LangevinIntegrator(300*unit.kelvin,
                                   1.0/unit.picoseconds,
                                   2.0*unit.femtoseconds)
integrator.setConstraintTolerance(0.00001)

simulation = app.Simulation(pdb.topology, system, integrator)
simulation.context.setPositions(pdb.positions)
""")
    with open(name, 'w', encoding='utf-8') as output:
        output.write(CODE)


def write_test_pdb(name):
    """
    This writes test.pdb with 2 waters into the current running directory.
    """
    with open(name, 'w', encoding='utf-8') as output:
        output.write(PDB_STRING)


class FakeOp:
    """Order parameter for testing the OpenMMEngine."""
    def __init__(self, values=None):
        self.values = values
        self.n = 0

    def calculate(self, _):
        """Calculate the fake order parameter."""
        value = self.values[self.n]
        self.n += 1
        return [value]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the methods in pyretis.setup.checker"""
import os
import logging
import unittest
from io import StringIO
from pyretis.engines.internal import VelocityVerlet
from pyretis.core.particles import ParticlesExt
from pyretis.core.system import System
from pyretis.inout.checker import (check_for_bullshitt,
                                   check_engine,
                                   check_ensemble,
                                   check_interfaces)
from pyretis.inout.settings import fill_up_tis_and_retis_settings
from unittest.mock import patch

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))

SETTINGS = {'simulation': {'task': 'retis',
                           'interfaces': [-1.0, -0.5, 0, 0.5, 1],
                           'exe_path': HERE,
                           'zero_ensemble': True,
                           'flux': True},
            'orderparameter': {'class': 'Position',
                               'dim': 'x',
                               'index': 0,
                               'periodic': False},
            'particles': {'type': 'internal'},
            'system': {'units': 'lj'},
            'engine': {'obj': VelocityVerlet(0.002)},
            'tis': {}}


def make_test_system(conf):
    """Just make a test system with particles."""
    system = System()
    system.particles = ParticlesExt(dim=3)
    system.particles.config = conf
    return system


class TestMethods(unittest.TestCase):
    """Test some of the methods from .checker."""

    def test_check_ensemble(self):
        """Test check_ensemble."""
        play_set = SETTINGS.copy()
        with self.assertRaises(ValueError) as err:
            self.assertFalse(check_ensemble(play_set))
        self.assertEqual('No ensemble in settings',
                         str(err.exception))

        fill_up_tis_and_retis_settings(play_set)
        play_set_base = play_set.copy()

        self.assertTrue(check_ensemble(play_set))

        play_set['simulation']['interfaces'][2] = 4
        with self.assertRaises(ValueError) as err:
            self.assertFalse(check_ensemble(play_set))
        self.assertEqual('The ensemble interface 0 '
                         'is not present in the simulation interface list',
                         str(err.exception))

        # Not sorted ensembles
        play_set = play_set_base.copy()
        play_set['ensemble'][0]['interface'] = 4
        play_set['ensemble'][1]['interface'] = 3
        with self.assertRaises(ValueError) as err:
            self.assertFalse(check_ensemble(play_set))
        self.assertTrue('NOT properly sorted' in str(err.exception))

        # Not declared ensembles
        play_set = play_set_base.copy()
        del play_set['ensemble'][0]['interface']
        with self.assertRaises(ValueError) as err:
            self.assertFalse(check_ensemble(play_set))
        self.assertTrue('without reference interface' in str(err.exception))

        play_set = play_set_base.copy()
        del play_set['ensemble']
        with self.assertRaises(ValueError) as err:
            self.assertFalse(check_ensemble(play_set))
        self.assertEqual('No ensemble in settings', str(err.exception))

    def test_check_engine(self):
        """Test check_engine."""
        play_set = SETTINGS
        play_set['particles'] = {'type': 'external'}
        play_set['engine'] = {'type': 'external',
                              'cp2k': 'cp2k',
                              'input_path': 'No thank you',
                              'timestep': 1,
                              'cp2k_format': 'xyz',
                              'subcycles': 1}
        self.assertTrue(check_engine(play_set))

        del play_set['engine']['input_path']
        self.assertFalse(check_engine(play_set))

        play_set['engine']['input_path'] = 'Hahaha'
        del play_set['engine']['cp2k_format']
        self.assertFalse(check_engine(play_set))

        play_set['engine']['cp2k_format'] = 'xyz'
        del play_set['engine']['cp2k']
        play_set['engine']['gmx'] = 'Buuuuu'
        self.assertFalse(check_engine(play_set))

    def test_check_interfaces(self):
        """Test check_interfaces."""
        settings = {'simulation': {'task': 'retis',
                                   'interfaces': [-1.0, -0.5, 0, 0.5, 1],
                                   'zero_ensemble': True,
                                   'flux': True}}

        self.assertTrue(check_interfaces(settings))
        settings['simulation']['zero_ensemble'] = False
        self.assertFalse(check_interfaces(settings))

        settings['simulation']['zero_ensemble'] = True
        settings['simulation']['interfaces'] = [1, 2]
        self.assertFalse(check_interfaces(settings))

        settings['simulation']['interfaces'] = [3, 2, 2.5]
        self.assertFalse(check_interfaces(settings))

        settings['simulation']['interfaces'] = [-1.0, -0.5, 0, 0.5, 1]
        self.assertTrue(check_interfaces(settings))

    def test_check_for_bullshitt(self):
        """Test that _check for bullshitt finds the inconsistent settings."""
        # Insufficient interfaces for TIS/RETIS:
        settings = {'simulation': {'task': 'tis', 'interfaces': [1]}}
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                check_for_bullshitt(settings)
        self.assertTrue('Insufficient number of interfaces for tis' in
                        str(err.exception))
        # No references:
        settings = {'simulation': {'task': 'tis', 'interfaces': [1, 2, 3]},
                    'ensemble': [{'gino': 'strada'}, {'interface': 1},
                                 {'interface': 3}]}
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                check_for_bullshitt(settings)
        self.assertTrue('An ensemble has been introduced without references'
                        ' (interface in ensemble settings)' in
                        str(err.exception))

        settings = {'simulation': {'task': 'tis', 'interfaces': [1, 2, 3]},
                    'ensemble': [{'ensemble_number': 0}, {'interface': 2},
                                 {'interface': 3}]}
        check_for_bullshitt(settings)

        # Not Sorted interfaces:
        settings = {'simulation': {'task': 'tis', 'interfaces': [2, 5, 1]}}
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                check_for_bullshitt(settings)
        self.assertTrue('Interface lambda positions in the simulation entry '
                        'are NOT sorted (small to large)' in
                        str(err.exception))

        # Wrong interfaces:
        settings = {'simulation': {'task': 'tis', 'interfaces': [1, 2, 3]},
                    'ensemble': [{'interface': 1}, {'interface': 2},
                                 {'interface': 4}]}
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                check_for_bullshitt(settings)
        self.assertTrue('in the simulation interface' in str(err.exception))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This script will help to copy trjectory files needed for restarting.

Here, we copy the last accepted path(s) from a previous simulation.
"""
import os
import sys
import shutil
import colorama
from pyretis.inout import print_to_screen
from pyretis.inout.archive import PathStorage
from pyretis.inout.settings import parse_settings_file
from pyretis.core.pathensemble import generate_ensemble_name


def read_path_file(filename):
    """Get index for last accepted path."""
    last_idx = None
    with open(filename, 'r') as fileh:
        for i, lines in enumerate(fileh):
            if i == 0:
                continue
            split = lines.strip().split()
            idx = int(split[0])
            status = split[7]
            if status == 'ACC':
                last_idx = idx
    return last_idx


def get_files_from_directory(source, target):
    """Investigate and copy an accepted path for the given ensemble."""
    print_to_screen(f'Checking directory:\xa0{source}',
                    level='info')
    path_file = os.path.join(source, 'pathensemble.txt')
    last_one = read_path_file(path_file)
    print_to_screen(f'Will use path no. {last_one}')

    # Copy the accepted last path from the traj archive.
    # This is the path where it is stored:
    traj_archive = os.path.join(
        source,
        'traj',  # Note, this is hardcoded in PyRETIS
        PathStorage.archive_acc,
        PathStorage.out_dir_fmt.format(last_one)
    )
    # We first copy the order, energy and traj files:
    for filei in ('order.txt', 'energy.txt', 'traj.txt'):
        src = os.path.join(traj_archive, filei)
        dest = os.path.join(target, filei)
        shutil.copy(src, dest)
    # Next, copy the actual trajectory files. Here we copy the files
    # corresponding to the last accepted trajectory found in the given
    # source. This trajectory is copied into a sub-directory called "accepted"
    # of the target directory.
    traj_archive_raw = os.path.join(traj_archive, PathStorage.archive_traj)
    for i in os.scandir(traj_archive_raw):
        if i.is_file:
            src = os.path.join(traj_archive_raw, i.name)
            target_dir = os.path.join(target, 'accepted')
            try:
                os.makedirs(target_dir)
            except FileExistsError:
                pass
            dest = os.path.join(target_dir, i.name)
            shutil.copy(src, dest)


def main(source, target):
    """Copy the files."""
    settings = parse_settings_file(os.path.join(source, 'retis.rst'))
    interf = settings['simulation']['interfaces']
    for i, _ in enumerate(interf):
        ens = generate_ensemble_name(i)
        target_ens = os.path.join(target, ens)
        source_ens = os.path.join(source, ens)
        try:
            os.makedirs(target_ens)
        except FileExistsError:
            pass
        for filei in ('pathensemble.txt', 'ensemble.restart'):
            src = os.path.join(source_ens, filei)
            dest = os.path.join(target_ens, filei)
            shutil.copy(src, dest)
        get_files_from_directory(source_ens, target_ens)


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main(sys.argv[1], sys.argv[2])



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Functions for generating reports.

The reports are useful for displaying results from the analysis.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_report_mdflux (:py:func:`.generate_report_mdflux`)
    Generate a report for a MD flux simulation.
"""
from pyretis.inout.report.markup import (generate_rst_table,
                                         generate_latex_table)
from pyretis.inout.formats.formatter import apply_format


__all__ = ['generate_report_mdflux']


def generate_report_mdflux(analysis, output='rst'):
    """Generate a report for a MD flux simulation.

    Parameters
    ----------
    analysis : dict
        This is the output from the analysis.
    output : string, optional
        This is the desired output format. It must match one of the
        formats defined in
        :py:const:`pyretis.inout.report.report.__TEMPLATES`. Default
        is 'rst' (reStructuredText).

    Returns
    -------
    out : dictionary
        The generated report as a string.

    """
    report = {'figures': {}, 'tables': {}}
    # generate some tables:
    cross_out = analysis['cross']['out']
    report['figures']['flux'] = analysis['cross']['figures']
    report['figures']['energy'] = analysis['energy']['figures']
    report['figures']['order'] = analysis['order']['figures']
    report['tables']['md-flux'] = _table_md_flux(cross_out, fmt=output)[1]
    report['tables']['md-cycles'] = _table_md_flux_cycles(cross_out,
                                                          fmt=output)[1]
    report['tables']['md-efficiency'] = _table_md_efficiency(cross_out,
                                                             fmt=output)[1]
    return report


def _table_md_efficiency(results, fmt='rst'):
    """Generate table with MD-flux results for efficiencies and correlations.

    Parameters
    ----------
    results : dict
        These are the results obtained by :py:func:`.analyse_flux`.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in the desired format which represents
        the table.

    """
    # table for interfaces:
    table = []
    for i, _ in enumerate(results['interfaces']):
        pmd = results['pMD'][i]
        teff = results['teffMD'][i]
        corr = results['corrMD'][i]
        prel = results['1-p'][i]
        row = [
            f'{i + 1:^10d}',
            apply_format(pmd, '{:^10.6f}'),
            apply_format(prel, '{:^10.6f}'),
            apply_format(teff, '{:^10.6f}'),
            apply_format(corr, '{:^10.6f}'),
        ]
        table.append(row)

    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Efficiency',
                                         ['Interface',
                                          r'$p_\text{MD}$',
                                          r'$\frac{1-p}{p}$',
                                          'Efficiency time', 'Correlation'],
                                         fixnum={1, 2, 3, 4})
    elif fmt in ['txt']:
        table_str = generate_rst_table(table, 'Efficiency',
                                       ['Interface',
                                        'p_MD',
                                        '(1-p_MD)/p_MD',
                                        'Efficiency time', 'Correlation'])
    else:
        table_str = generate_rst_table(table, 'Efficiency',
                                       ['Interface',
                                        r':math:`p_\text{MD}`',
                                        r':math:`\frac{1-p}{p}`',
                                        'Efficiency time', 'Correlation'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_md_flux_cycles(results, fmt='rst'):
    """Generate a table for the MD-flux results for cycle numbers.

    The table will display the number of steps in state A, state B,
    overall state A and B and the total number of MD cycles.

    Parameters
    ----------
    results : dict
        These are the results obtained by :py:func:`.analyse_flux`.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in the desired format which represents
        the table.

    """
    # table for interfaces:
    table = [
        ['A', f'{results["times"]["A"]:8d}'],
        ['B', f'{results["times"]["B"]:8d}'],
        ['overall A', f'{results["times"]["OA"]:8d}'],
        ['overall B', f'{results["times"]["OB"]:8d}'],
        ['Total cycles', f'{results["totalcycle"]:8d}'],
    ]
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Cycles spent in state',
                                         ['State', 'Cycles'],
                                         fixnum={2})
    else:
        table_str = generate_rst_table(table, 'Cycles spent in state',
                                       ['State', 'Cycles'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_md_flux(results, fmt='rst'):
    """Generate the table for the MD-flux results.

    Parameters
    ----------
    results : dict
        These are the results obtained by :py:func:`.analyse_flux`.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in the desired format which represents
        the table.

    """
    # table for interfaces:
    table = []
    for i, idet in enumerate(results['interfaces']):
        flux = results['runflux'][i][-1]
        error = results['errflux'][i][2]
        relerror = results['errflux'][i][4]
        row = [
            f'{i + 1:^4d}',
            apply_format(idet, '{:^8.4f}'),
            apply_format(flux, '{:^10.6f}'),
            apply_format(error, '{:^10.6f}'),
            apply_format(relerror, '{:^10.6f}'),
            f'{results["ncross"][i]:^8d}',
            f'{results["neffcross"][i]:^8d}',
            apply_format(results['neffc/nc'][i], '{:^10.6f}'),
            apply_format(results['cross_time'][i], '{:^10.6f}'),
        ]
        table.append(row)
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Flux for interfaces',
                                         ['Int.', 'Position', 'Flux / units',
                                          'Error', 'Rel. error', 'Ncross',
                                          'Neffcross', 'Neffcross/Ncross',
                                          'Steps per cross'],
                                         fixnum={2, 3, 4, 5, 6, 7, 8})
    else:
        table_str = generate_rst_table(table, 'Flux for interfaces',
                                       ['Int.', 'Position', 'Flux / units',
                                        'Error', 'Rel. error', 'Ncross',
                                        'Neffcross', 'Neffcross/Ncross',
                                        'Steps per cross'])
    table_txt = '\n'.join(table_str)
    return table, table_txt



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for the xyz in/out module."""
import logging
import unittest
import tempfile
import filecmp
import os
import gzip
import numpy as np
from pyretis.inout.formats.xyz import (
    format_xyz_data,
    read_xyz_file,
    convert_snapshot,
    write_xyz_file,
    write_xyz_trajectory,
    reverse_xyz_file,
    xyz_merge,
    txt_to_xyz,
)


logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


class XYZIOTest(unittest.TestCase):
    """Test the xyz format reading & writing."""

    def test_format_data(self):
        """Test that we format data as expected."""
        correct_xyz = [
            ['2', 'PyRETIS XYZ writer',
             'X         1.000000000     2.000000000     3.000000000',
             'X         4.000000000     5.000000000     6.000000000'],
            ['2', 'Custom header',
             'X         1.000000000     2.000000000     3.000000000',
             'X         4.000000000     5.000000000     6.000000000'],
            ['2', 'PyRETIS XYZ writer',
             'A         1.000000000     2.000000000     3.000000000',
             'B         4.000000000     5.000000000     6.000000000'],
            ['2', 'PyRETIS XYZ writer',
             'X          1.0      2.0      3.0',
             'X          4.0      5.0      6.0'],
            ['2', 'PyRETIS XYZ writer',
             ('X         1.000000000     2.000000000     3.000000000'
              '     9.000000000     8.000000000     7.000000000'),
             ('X         4.000000000     5.000000000     6.000000000'
              '     6.000000000     5.000000000     4.000000000')],
        ]
        pos = np.array([[1., 2., 3.], [4., 5., 6.]])
        vel = np.array([[9., 8., 7.], [6., 5., 4.]])
        cases = [
            {'pos': pos},
            {'pos': pos, 'header': 'Custom header'},
            {'pos': pos, 'names': ['A', 'B']},
            {'pos': pos, 'fmt': '{0:5s} {1:8.1f} {2:8.1f} {3:8.1f}'},
            {'pos': pos, 'vel': vel},
        ]
        for case, correct in zip(cases, correct_xyz):
            formatted = format_xyz_data(
                case.get('pos', None),
                vel=case.get('vel', None),
                names=case.get('names', None),
                header=case.get('header', None),
                fmt=case.get('fmt', None),
            )
            for i, j in zip(formatted, correct):
                self.assertEqual(i, j)

    def test_read_file(self):
        """Test that we can read a single snapshot."""
        correct_data = [
            {'box': np.array([1., 2., 3.]),
             'xyz': np.array([[0., 0., 0.], [0.5, 0.5, 0.5], [0.5, 0.5, 0.],
                              [0.5, 0., 0.5], [0., 0.5, 0.5]]),
             'vel': np.zeros((5, 3)),
             'names': ['Ba', 'Hf', 'O', 'O', 'O']},
            {'box': np.array([1., 2., 3., 4., 5., 6.]),
             'xyz': np.array([[1., 2., 3.], [7., 8., 9.], [0.1, 0.2, 0.3]]),
             'vel': np.array([[4., 5., 6.], [10., 11., 12.], [0.6, 0.7, 0.8]]),
             'names': ['X', 'Y', 'Z']},
        ]
        infiles = ('config_box.xyz', 'config_vel.xyz')
        for files, correct in zip(infiles, correct_data):
            infile = os.path.join(HERE, files)
            snapshot = read_xyz_file(infile)
            box, xyz, vel, names = convert_snapshot(next(snapshot))
            data = (box, xyz, vel)
            var = ('box', 'xyz', 'vel')
            for i, j in zip(data, var):
                self.assertTrue(np.allclose(i, correct[j]))
            for i, j in zip(names, correct['names']):
                self.assertEqual(i, j)

    def test_xyz_file(self):
        """Test that we can write an xyz-file."""
        infile = os.path.join(HERE, 'config_read.xyz')
        snapshot = next(read_xyz_file(infile))
        _, xyz, vel, names = convert_snapshot(snapshot)
        header = snapshot['header']
        with tempfile.NamedTemporaryFile() as temp:
            write_xyz_file(temp.name, xyz, vel=vel, names=names,
                           header=header)
            temp.flush()
            compare = filecmp.cmp(temp.name, infile)
            self.assertTrue(compare)

    def test_write_trajectory(self):
        """Test that we can write a xyz-trajectory."""
        posin = [np.ones((2, 3))*1.5, np.ones((2, 3))*2.]
        velin = [np.ones((2, 3)), np.ones((2, 3))*2.]
        namesin = ['A', 'B']
        boxin = [[3., 3., 3.], [5., 5., 5., 5., 5., 5.]]
        step = 0
        headers = [
            '# Step: 0 Box:    3.0000    3.0000    3.0000',
            ('# Box:    5.0000    5.0000    5.0000'
             '    5.0000    5.0000    5.0000'),
        ]
        with tempfile.NamedTemporaryFile() as temp:
            for i, j, k in zip(posin, velin, boxin):
                write_xyz_trajectory(temp.name, i, j, namesin,
                                     k, step=step, append=True)
                step = None
            temp.flush()
            for i, snapshot in enumerate(read_xyz_file(temp.name)):
                box, xyz, vel, names = convert_snapshot(snapshot)
                self.assertTrue(np.allclose(xyz, posin[i]))
                self.assertTrue(np.allclose(vel, velin[i]))
                self.assertTrue(np.allclose(box, boxin[i]))
                self.assertEqual(snapshot['header'], headers[i])
                for j, k in zip(namesin, names):
                    self.assertEqual(j, k)

    def test_reverse_xyz_file(self):
        """Test that we can reverse an xyz-file."""
        infile = os.path.join(HERE, 'backward.xyz')
        with tempfile.NamedTemporaryFile() as temp:
            reverse_xyz_file(infile, temp.name)
            temp.flush()
            correct = reversed([snap for snap in read_xyz_file(infile)])
            for snapshot, corr in zip(read_xyz_file(temp.name), correct):
                self.assertEqual(snapshot['header'], corr['header'])
                for key, val in snapshot.items():
                    if key in ('header', 'atomname'):
                        self.assertEqual(val, corr[key])
                    else:
                        self.assertTrue(np.allclose(val, corr[key]))

    def test_reverse_larger_file(self):
        """Test that we can reverse a larger xyz-file."""
        infile = os.path.join(HERE, 'traj.xyz')
        with tempfile.NamedTemporaryFile() as temp:
            reverse_xyz_file(infile, temp.name)
            temp.flush()
            for i, snapshot in enumerate(read_xyz_file(temp.name)):
                step = int(snapshot['header'].split()[2])
                self.assertEqual(i, step-1)
                posx = np.ones(3) * (i + 1)
                self.assertTrue(np.allclose(snapshot['x'], posx))

    def test_merge(self):
        """Test that we can merge forward and backward xyz-files."""
        backward = os.path.join(HERE, 'backward.xyz')
        forward = os.path.join(HERE, 'forward.xyz')
        merged_correct = os.path.join(HERE, 'merged.xyz')
        with tempfile.NamedTemporaryFile() as temp:
            xyz_merge(backward, forward, temp.name)
            temp.flush()
            compare = filecmp.cmp(temp.name, merged_correct)
            self.assertTrue(compare)

    def test_txt_convert(self):
        """Test that we can convert from the internal txt format."""
        infile = os.path.join(HERE, 'traj.txt.gz')
        output_name = os.path.join(HERE, 'DELETE')
        out = []
        correct = ['correct-000-ACC.xyz.gz', 'correct-000-ACC.xyz.gz']
        with tempfile.NamedTemporaryFile() as temp:
            with open(temp.name, 'wb') as output:
                with gzip.open(infile, 'rb') as gfile:
                    output.write(gfile.read())
            temp.flush()
            out = txt_to_xyz(temp.name, output_name, None,
                             selection='ACC', nzero=3)
        for files, correctf in zip(out, correct):
            correct_name = os.path.join(HERE, correctf)
            with tempfile.NamedTemporaryFile() as temp:
                with open(temp.name, 'wb') as output:
                    with gzip.open(correct_name, 'rb') as gfile:
                        output.write(gfile.read())
                temp.flush()
                compare = filecmp.cmp(temp.name, files)
                self.assertTrue(compare)
            os.remove(files)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of creating a simple order parameter."""
from pyretis.core import System, Particles
from pyretis.orderparameter import Position

position = Position(0, dim='x')

# Define a simple system for testing:
system = System()
system.particles = Particles(dim=3)
system.add_particle([1.0, 2.0, 3.0], mass=1.0, name='Ar', ptype=0)
print('Order =', position.calculate(system))



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of force field classes and potential functions.

In PyRETIS a force field is just a collection of potential functions
with some parameters. This module defines the force field and the
potential functions that can be used to build up force fields.

Package structure
-----------------

Modules
~~~~~~~

forcefield.py (:py:mod:`pyretis.forcefield.forcefield`)
    Defines the force field class (:py:class:`.ForceField`) which can
    be used to represent a generic force field.

potential.py (:py:mod:`pyretis.forcefield.potential`)
    Defines the generic potential function class
    (:py:class:`.PotentialFunction`) which is sub-classed in other
    potential functions.

factory.py (:py:mod:`pyretis.forcefield.factory`)
    Defines a method for creating potentials from input settings.

Sub-packages
~~~~~~~~~~~~

potentials (:py:mod:`pyretis.forcefield.potentials`)
    Definition of potential functions for force fields.


Important classes defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ForceField (:py:class:`.ForceField`)
    A class representing a general force field.

PotentialFunction (:py:class:`.PotentialFunction`)
    A class representing a general potential function.
"""
from .forcefield import ForceField
from .potential import PotentialFunction
from . import potentials



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file defines the order parameter used for the WCA example."""
import logging
import numpy as np
from pyretis.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class OrderXY(OrderParameter):
    """OrderXY(OrderParameter).

    This class defines a 2D order parameter such that x + y = constant.

    Attributes
    ----------
    index : integer
        This selects the particle to use for the order parameter.
    inter_a : float
        An interface such that we are in state A for postions < inter_a.
    inter_b : float
        An interface such that we are in state B for postions > inter_b.
    energy_a : float
        An energy such that we are in state A for potential energy < energy_a.
    energy_b : float
        An energy such that we are in state A for potential energy < energy_b.

    """

    def __init__(self, index, inter_a, inter_b, energy_a, energy_b):
        """Set up the order parameter.

        Parameters
        ----------
        index : tuple of ints
            This is the indices of the atom we will use the position of.
        inter_a : float
            An interface such that we are in state A for postions < inter_a.
        inter_b : float
            An interface such that we are in state B for postions > inter_b.
        energy_a : float
            An energy such that we are in state A for
            potential energy < energy_a.
        energy_b : float
            An energy such that we are in state A for
            potential energy < energy_b.

        """
        super().__init__(description='2D->1D projection')
        self.index = index
        x_1 = 0.2
        y_1 = 0.4
        x_0 = -0.2
        y_0 = -0.4
        self.x_0 = x_0
        self.y_0 = y_0
        self.origin = np.array([x_0, y_0])
        self.vec = np.array([x_1 - x_0, y_1 - y_0])
        self.vec /= np.sqrt(np.dot(self.vec, self.vec))
        self.inter_a = inter_a
        self.inter_b = inter_b
        self.energy_a = energy_a
        self.energy_b = energy_b

    def calculate(self, system):
        """Calculate the order parameter.

        Here, the order parameter is just the distance between two
        particles.

        Parameters
        ----------
        system : object like :py:class:`.System`
            This object is used for the actual calculation, typically
            only `system.particles.pos` and/or `system.particles.vel`
            will be used. In some cases `system.forcefield` can also be
            used to include specific energies for the order parameter.

        Returns
        -------
        out : float
            The order parameter.

        """
        pos = system.particles.pos[self.index]
        vec = np.array([pos[0] - self.x_0, pos[1] - self.y_0])
        proj = np.dot(vec, self.vec)
        if proj < self.inter_a:
            if system.particles.vpot > self.energy_a:
                proj = self.inter_a
        elif proj > self.inter_b:
            if system.particles.vpot > self.energy_b:
                proj = self.inter_b
        return [proj]



"""Run linting for a set of source files.

Here, we will also give some over-all evaluation of the results.
The exit code of this script will be != 0 if one or more files
have a score below a pre-set value or if the changes are "too big".
"""
import argparse
import colorama
import io
import pathlib
import re
import sys
from colorama import Fore
from operator import itemgetter
from pathlib import Path
from pylint import lint
from pylint.reporters.text import TextReporter

SCORE_THRESHOLD = 6.7


def look_for_source_files(rootdir, skip=None):
    """Find .py files by walking the given directory."""
    root = pathlib.Path(rootdir)
    glob = sorted(root.rglob('*.py'))
    if skip is None:
        return glob
    skip_dirs = [pathlib.Path(i) for i in skip]
    return [
        i for i in glob if not any(parent in skip_dirs for parent in i.parents)
    ]


def lint_file(filename, reg_score, reg_warning, reg_error, white_list=None):
    """Run linting for a single file."""
    if white_list is None:
        cmd = str(Path(filename))
    else:
        cmd = [str(Path(filename)),
               f"--ignored-modules={','.join(white_list)}"]

    buf = io.StringIO()

    lint.Run(cmd, reporter=TextReporter(buf), exit=False)

    txt = buf.getvalue()

    warnings = reg_warning.findall(txt)
    errors = reg_error.findall(txt)
    result = reg_score.findall(txt)

    score, delta = process_result(result)

    return score, delta, warnings, errors


def run_linting(files, white_list=None):
    """Run pylint for a set of files, individually."""
    reg_score = re.compile(r'-?\d+\.\d+\/\d+')
    reg_warning = re.compile(r'.+warning \(W.+')
    reg_error = re.compile(r'.+error \(E.+')
    results = {
        'all': [],
        'below': [],
        'big-change': [],
        'change': [],
        'warnings': [],
        'errors': []
        }
    print(Fore.BLUE + 'Running pylint...')
    for i, filei in enumerate(files):
        printed = False
        if i % 10 == 0:
            print(Fore.BLUE + '{} files remaining...'.format(len(files) - i))
        score, delta, warnings, errors = lint_file(
            filei,
            reg_score,
            reg_warning,
            reg_error,
            white_list=white_list,
        )
        if score is None:
            print(Fore.YELLOW + 'Skipping file: {}'.format(filei))
            continue

        if delta is not None:
            results['change'].append((filei, score, delta, abs(delta)))
            if delta < -0.5:
                results['big-change'].append((filei, score, delta))
                if not printed:
                    print(Fore.RED + '{} {} {}'.format(filei, score, delta))
                    printed = True

        if score < SCORE_THRESHOLD:
            results['below'].append((filei, score, delta))
            if not printed:
                print(Fore.RED + '{} {} {}'.format(filei, score, delta))
                printed = True

        if not printed:
            print(filei, score, delta)

        if warnings:
            results['warnings'].append((filei, score, delta))
            print_list(warnings, txt='warnings', color=Fore.YELLOW)
        if errors:
            results['errors'].append((filei, score, delta))
            print_list(errors, txt='errors', color=Fore.RED)

        results['all'].append((filei, score, delta))
    return results


def process_result(result):
    """Do a simple evaluation of the score."""
    try:
        score = float(result[0].split('/')[0])
    except IndexError:
        return None, None
    try:
        delta = score - float(result[1].split('/')[0])
    except IndexError:
        delta = None
    return score, delta


def print_list(list_to_print, txt='warnings', color=Fore.YELLOW):
    """Print out the contents in a list with some colour and text."""
    if list_to_print:
        print(color + '    Found {}:'.format(txt))
        for i in list_to_print:
            print(color + '    {}'.format(i.strip()))


def print_below(results):
    """Print the below results if any."""
    if results['below']:
        fmt = '    {}: {} (delta: {})'
        print()
        print(Fore.RED + 'The following files *MUST BE FIXED*:')
        for i in results['below']:
            print(fmt.format(*i))
        return 1
    return 0


def print_big_change(results):
    """Print big results if any."""
    if results['big-change']:
        fmt = '    {}: {} (delta: {})'
        print()
        print(Fore.RED +
              'The following files had a *TOO LARGE* negative change:')
        for i in results['big-change']:
            print(fmt.format(*i))
        return 1
    return 0


def print_all_warnings(results):
    """Print files with warnings if any."""
    if results['warnings']:
        fmt = '    {}'
        print()
        print(Fore.YELLOW +
              'The following files gave warnings:')
        for i in results['warnings']:
            print(fmt.format(i[0]))
        return 1
    return 0


def print_all_errors(results):
    """Print files with errors if any."""
    if results['errors']:
        fmt = '    {}'
        print()
        print(Fore.RED +
              'The following files have errors:')
        for i in results['errors']:
            print(fmt.format(i[0]))
        return 1
    return 0


def evaluate_results(results):
    """Do a simple evaluation of the results."""
    fmt = '    {}: {} (delta: {})'
    # First print out 10 (if possible) files with lowerest scores:
    maxlen = min(10, len(results['all']))
    txt = []
    for i in sorted(results['all'], key=itemgetter(1))[:maxlen]:
        if i[1] < 10.0:
            txt.append(fmt.format(*i))
    print()
    if txt:
        print(Fore.BLUE + 'The lowest scores:')
        for i in txt:
            print(i)
    else:
        print(Fore.GREEN + 'No scores below 10! Good job :-)')

    # Print out the 10 biggest changes:
    maxlen = min(10, len(results['change']))
    if maxlen > 0:
        txt = []
        for i in sorted(results['change'], key=itemgetter(3),
                        reverse=True)[:maxlen]:
            if abs(i[3]) > 0.0:
                txt.append(fmt.format(*i))
        print()
        if txt:
            print(Fore.BLUE + 'The biggest changes:')
            for i in txt:
                print(i)
        else:
            print()
            print(Fore.BLUE + 'No changes found.')

    fail1 = print_below(results)
    fail2 = print_big_change(results)
    print_all_warnings(results)
    fail3 = print_all_errors(results)
    return fail1 + fail2 + fail3


def create_argument_parser():
    """Create a argument parser for the linting tool."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i',
        '--input',
        nargs='+',
        help='Input directories to search for .py files.',
        required=True,
    )
    parser.add_argument(
        '-s',
        '--skip',
        nargs='+',
        help='Subdirectories to skip.',
        required=False,
    )
    parser.add_argument(
        '-w',
        '--white-list',
        nargs='+',
        help=(
            'Python modules to white-list, please see the pylint '
            'documentation on the option: '
            '"extension-pkg-whitelist"'
        ),
        required=False,
    )
    return parser


def main(args):
    """Run the analysis."""
    return_code = 0
    for startdir in args.input:
        print(Fore.GREEN + '\nRunning in: "{}"'.format(startdir))
        files = look_for_source_files(startdir, skip=args.skip)
        if files:
            lint_result = run_linting(files, white_list=args.white_list)
            result = evaluate_results(lint_result)
            return_code += result
        else:
            print(Fore.RED + 'No files found in {}!'.format(startdir))
            return_code += 1
    sys.exit(return_code)


if __name__ == '__main__':
    colorama.init(autoreset=True)
    PARSER = create_argument_parser()
    main(PARSER.parse_args())



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A double pendulum potential."""
import logging
import numpy as np
from pyretis.forcefield.potential import PotentialFunction
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class DoublePendulum(PotentialFunction):
    r"""DoublePendulum(PotentialFunction).

    Attributes
    ----------
    l1 : float
        Length of first wire.
    l2 : float
        Length of second wire.
    g : float
        The acceleration of gravity.

    """

    def __init__(self, l1=1.0, l2=1.0, g=1.0, desc='2D double pendulum'):
        """Set up the double pendulum potential.

        Parameters
        ----------
        l1 : float, optional
            Length of first wire.
        l1 : float, optional
            Length of second wire.
        g : float, optional
            The acceleration of gravity.
        desc : string, optional
            Description of the force field.

        """
        super().__init__(dim=2, desc=desc)
        self.params = {'g': g, 'l1': l1, 'l2': l2}
        self.l1 = l1
        self.l2 = l2
        self.g = g

    def potential(self, system):
        """Evaluate the potential energy.

        Parameters
        ----------
        system : object like `System`
            The system we evaluate the potential for. Here, we
            make use of the positions only.

        Returns
        -------
        out : float
            The potential energy.

        """
        costheta1 = np.cos(system.particles.pos[0][0])
        costheta2 = np.cos(system.particles.pos[1][0])
        m1 = system.particles.mass[0][0]
        m2 = system.particles.mass[1][0]
        return -self.g * ((m1 + m2) * self.l1 * costheta1 +
                          m2 * self.l2 * costheta2)

    def force(self, system):
        """Evaluate forces for the double pendulum.

        Parameters
        ----------
        system : object like `System`
            The system we evaluate the potential for. Here, we
            make use of the positions only.

        Returns
        -------
        out[0] : numpy.array
            The calculated force.
        out[1] : numpy.array
            The virial, currently not implemented for this potential.

        """
        g = self.g
        l1 = self.l1
        l2 = self.l2
        pos = system.particles.pos
        vel = system.particles.vel
        m1 = system.particles.mass[0][0]
        m2 = system.particles.mass[1][0]
        theta1 = pos[0][0]
        theta2 = pos[1][0]
        dtheta1 = vel[0][0]
        dtheta2 = vel[1][0]
        dth = theta1 - theta2
        M = m2 / (m1 + m2)
        ll = l2 / l1
        wsq = g / l1
        denom = ll*(1.0 - M*np.cos(dth)**2)
        atheta1 = wsq*ll*(-np.sin(theta1) + M*np.cos(dth)*np.sin(theta2))
        atheta1 -= M*ll*np.sin(dth)*(dtheta1**2*np.cos(dth) +
                                     ll*dtheta2**2)
        atheta1 /= denom

        atheta2 = wsq*np.cos(dth)*np.sin(theta1) - wsq*np.sin(theta2)
        atheta2 += (dtheta1**2 + M*ll*dtheta2**2*np.cos(dth))*np.sin(dth)
        atheta2 /= denom

        forces = np.zeros((2, self.dim))
        forces[0, 0] = atheta1
        forces[1, 0] = atheta2
        virial = np.zeros((self.dim, self.dim))  # just return zeros here
        return forces, virial

    def potential_and_force(self, system):
        """Evaluate the potential and the force.

        Parameters
        ----------
        system : object like `System`
            The system we evaluate the potential for. Here, we
            make use of the positions only.

        Returns
        -------
        out[0] : float
            The potential energy as a float.
        out[1] : numpy.array
            The force as a numpy.array of the same shape as the
            positions in `particles.pos`.
        out[2] : numpy.array
            The virial, currently not implemented for this potential.

        """
        forces, virial = self.force(system)
        pot = self.potential(system)
        return pot, forces, virial



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Version information for PyRETIS.

This file is generated by PyRETIS (``setup_version.py``)
"""
SHORT_VERSION = '3.0.1'
VERSION = '3.0.1'
FULL_VERSION = '3.0.1'
GIT_REVISION = '3135263ee53348bf3aa4bf685622f988dcc881f5'
GIT_VERSION = '3.0.1'
RELEASE = True

if not RELEASE:
    VERSION = GIT_VERSION

MIN_PYTHON_VERSION = (3, 9)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Methods for analysis of crossings for flux data.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

analyse_flux (:py:func:`.analyse_flux`)
    Run analysis for simulation flux data. This will calculate
    the initial flux for a simulation.
"""
import numpy as np
from numpy import divide  # pylint: disable=no-name-in-module
from pyretis.analysis.analysis import running_average, block_error_corr
from pyretis.core.path import check_crossing

np.set_printoptions(legacy='1.21')

__all__ = ['analyse_flux', 'find_crossings']


def analyse_flux(fluxdata, settings):
    """Run the analysis on the given flux data.

    This will run the flux analysis and collect the results into a
    structure which is convenient for plotting and reporting the
    results.

    Parameters
    ----------
    fluxdata : list of tuples of integers
        This array contains the data obtained from a MD simulation
        for the fluxes.
    settings : dict
        This dict contains the settings for the analysis. Note that
        this dictionary also needs some settings from the simulation,
        in particular the number of cycles, the interfaces and
        information about the time step.

    Returns
    -------
    results : dict
        This dict contains the results from the flux analysis.
        The keys are defined in the `results` variable.

    """
    end_step = settings['simulation']['endcycle']
    time_subcycles = settings['engine'].get('subcycles', 1)
    time_step = settings['engine']['timestep']*time_subcycles
    interfaces = list(settings['simulation']['interfaces'])
    results = {'eff_cross': [],  # effective crossings times
               'ncross': None,  # number of crossings
               'neffcross': [],  # number of effective crossings
               'times': {},  # time spent in the different states
               'flux': [],  # store raw flux data
               'runflux': [],  # running average of flux
               'errflux': [],  # block error analysis
               'interfaces': interfaces,
               'totalcycle': end_step,  # store total number of cycles
               'cross_time': [],  # steps per crossing
               'neffc/nc': [],  # Effective crossings per crossing
               'pMD': [],  #
               '1-p': [],  #
               'teffMD': [],  #
               'corrMD': []}  #
    if not fluxdata:
        return results
    ret = _effective_crossings(fluxdata, len(results['interfaces']), end_step)
    results['eff_cross'] = ret[0]
    results['ncross'] = ret[1]
    results['neffcross'] = ret[2]
    results['times'] = ret[3]
    analysis = settings['analysis']
    for i in range(len(results['interfaces'])):
        time, ncross, flux = _calculate_flux(results['eff_cross'][i],
                                             results['times']['OA'],
                                             analysis['skipcross'],
                                             time_step)
        results['flux'].append(np.column_stack((time, ncross, flux)))
        # now it's also a good time to obtain running averages etc.:
        results['runflux'].append(running_average(flux))
        block_error = block_error_corr(flux,
                                       maxblock=analysis['maxblock'],
                                       blockskip=analysis['blockskip'])
        results['errflux'].append(block_error)

    # do some additional statistics:
    results['cross_time'] = [divide(float(end_step), float(neff))
                             for neff in results['neffcross']]

    results['neffc/nc'] = [divide(float(neff), float(ncr)) for neff, ncr
                           in zip(results['neffcross'], results['ncross'])]
    for flux, error in zip(results['runflux'], results['errflux']):
        results['pMD'].append(flux[-1] * time_step)
        results['1-p'].append(divide(float(1.0 - results['pMD'][-1]),
                                     results['pMD'][-1]))
        results['teffMD'].append(end_step * error[4]**2)
        results['corrMD'].append(
            divide(results['teffMD'][-1], results['1-p'][-1])
        )
    return results


def _effective_crossings(fluxdata, nint, end_step):
    """Analyse flux data and obtain effective crossings.

    Parameters
    ----------
    fluxdata : list of tuples of ints
        The list contains the data obtained from a ``md-flux``
        simulation.
    nint : int
        The number of interfaces used.
    end_step : int
        This is the last step done in the simulation.

    Returns
    -------
    eff_cross : list of lists
        `eff_cross[i]` is the effective crossings times for
        interface `i`.
    ncross : list of ints
        `ncross[i]` is the number of crossings for interface `i`.
    neffcross : list of ints
        `neffcross[i]` is the number of effective crossings for
        interface `i`.
    time_in_state : dict
        The time spent in the different states which are labeled with
        the keys 'A', 'B', 'OA', 'OB'. 'O' is taken to mean the
        'overall' state.

    Note
    ----
    We do here `intf - 1`. This is just to be compatible with the old
    FORTRAN code where the interfaces are numbered 1, 2, 3 rather than
    0, 1, 2. If this is to be changed in the future the `-1` can just
    be removed. Such a change will also require changes to the
    formatter for flux data.

    """
    # First line is used to determine if we start in B or A
    overallstate_a = not (fluxdata[0][1] == 2 and fluxdata[0][2] < 0)
    firstcross = [True] * nint
    ncross = [0] * nint
    neffcross = [0] * nint
    eff_cross = [[] for _ in range(nint)]
    end = {'A': 0, 'B': 0, 'OA': 0, 'OB': 0}
    start = {'A': 0, 'B': 0, 'OA': 0, 'OB': 0}
    time_in_state = {'A': 0, 'B': 0, 'OA': 0, 'OB': 0}
    time, intf, sign = None, None, None
    for (time, intf, sign) in fluxdata:
        if sign > 0:  # positive direction
            if intf - 1 == 0:  # moving out of A
                end['A'] = time
                time_in_state['A'] += (end['A'] - start['A'])
            elif intf - 1 == 2:  # moving into B
                start['B'] = time
                if overallstate_a:  # if we came from A
                    end['OA'] = time
                    start['OB'] = time
                    time_in_state['OA'] += (end['OA'] - start['OA'])
                    overallstate_a = False
            ncross[intf - 1] += 1
            if firstcross[intf - 1]:
                firstcross[intf - 1] = False
                neffcross[intf - 1] += 1
                eff_cross[intf - 1].append((time - time_in_state['OB'], time))
        elif sign < 0:
            if intf - 1 == 0:  # moving into A
                firstcross = [True] * nint
                start['A'] = time
                if not overallstate_a:  # if we came from B
                    end['OB'] = time
                    start['OA'] = time
                    time_in_state['OB'] += (end['OB'] - start['OB'])
                    overallstate_a = not overallstate_a
            elif intf - 1 == 2:  # moving out of B
                end['B'] = time
                time_in_state['B'] += (end['B'] - start['B'])
    # Now, just add up the remaining:
    state = 'OA' if overallstate_a else 'OB'
    time_in_state[state] += (end_step - start[state])
    if intf - 1 == 0 and sign < 0:
        # Note that the sign < 0 works for sign=None
        time_in_state['A'] += (end_step - start['A'])
    elif intf - 1 == 2 and sign > 0:
        # Note that the sign > 0 works for sign=None
        time_in_state['B'] += (end_step - start['B'])
    return eff_cross, ncross, neffcross, time_in_state


def _calculate_flux(effective_cross, time_in_state, time_window, time_step):
    """Calculate the flux in different time windows.

    Parameters
    ----------
    effective_cross : list
        The number of effective crossings, obtained from
        ``_effective_crossings``.
    time_in_state : int
        Time spent in over-all state ``A``.
    time_window : int
        This is the time window we consider for calculating the flux.
    time_step : float
        This is the time-step for the simulation.

    Returns
    -------
    time : np.array
        The times for which we have calculated the flux.
    ncross : np.array
        The number of crossings within a time window.
    flux : np.array
        The flux within a time window.

    """
    max_windows = int(1.0 * time_in_state / time_window)
    ncross = np.zeros(max_windows, dtype=int)
    for crossing in effective_cross:
        idx = int(np.floor((crossing[0] - 0.0) / time_window))
        idx = min(idx, max_windows - 1)
        ncross[idx] += 1
    flux = (1.0 * ncross) / (time_step * time_window)
    time = np.arange(1, max_windows+1) * time_window
    return time, ncross, flux


def find_crossings(order, interfaces):
    """Find crossings with interfaces for given order parameter data.

    Parameters
    ----------
    order : numpy.array (1D)
        Order parameters, as a function of time.
    interfaces : list of floats
        The interfaces for which we will investigate crossings.

    Returns
    -------
    out : list of tuple
        Each tuple contains the crossings on the following form:
        (step, interface-number, direction), where direction = '+' if
        the interface was crossed while moving to the right and '-' if
        the movement was towards the left.

    """
    leftside_prev = None  # previous
    cross = []
    for step, orderi in enumerate(order):
        leftside, ncross = check_crossing(step, orderi,
                                          interfaces, leftside_prev)
        leftside_prev = list(leftside)
        cross.extend(ncross)
    return cross



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""PyRETIS.

This file is part of PyRETIS - a simulation package for rare events.

Copyright (c) 2026, PyRETIS Development Team

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


PyRETIS documentation
---------------------

The documentation for PyRETIS is available either doc-strings provided
with the code and from `the PyRETIS homepage <http://www.pyretis.org>`_.

PyRETIS sub-packages
--------------------

analysis (:py:mod:`pyretis.analysis`)
    Analysis tools for calculating crossing probabilities, rates etc.

core (py:mod:`pyretis.core`)
    Core classes and functions for running the rare event simulations.
    This includes classes defining the system, particles, simulations
    etc.

engines (:py:mod:`pyretis.engines`)
    A package defining engines which can be used to
    evolve the dynamics/solve Newton's equations of motion in time.

forcefield (:py:mod:`pyretis.forcefield`)
    This package defines force fields and potentials functions.

inout (:py:mod:`pyretis.inout`)
    This package defines the input-output operations for PyRETIS.
    This includes generating output from the analysis and reading
    input-files etc.

orderparameter (:py:mod:`pyretis.orderparameter`)
    Definition of classes for order parameters. Defines the base class
    for order parameters.

testing (:py:mod:`pyretis.testing`)
    This package defines common methods which are used in testing.

tools (:py:mod:`pyretis.tools`)
    This package defines some functions which can be useful for
    setting up simple systems, for example, functions for generating
    lattices.

pyvisa (:py:mod:`pyretis.pyvisa`)
    This package contains the compressor, the gui and analysis of
    the package PyVisA. This is component nearly independent from
    PyRETIS and has different prerequisites.
    Note: if PyQt5 is not installed, all PyVisA functions that
    are connected to the GUI are not usable. We thus only check here
    if PyQt5 is present.

"""
# PyRETIS imports:
from .version import VERSION as __version__
from .version import MIN_PYTHON_VERSION
from . import analysis
from . import core
from . import engines
from . import forcefield
from . import info
from . import initiation
from . import inout
from . import orderparameter
from . import setup
from . import testing
from . import tools
# PyVisA imports:
from .pyvisa import HAS_PYVISA_REQ
if HAS_PYVISA_REQ:
    from . import pyvisa  # pragma: no cover



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Methods that might be useful for testing.

This module defines generic methods for testing.

"""
import os


def search_for_files(rootdir, match=None):
    """Find files by walking the given directory.

    Parameters
    ----------
    rootdir : string
        The path where we will search from.
    match : string, optional
        If given, the method will only return files that
        are equal to the given match.

    Return
    ------
    out : list of strings
        The paths of the found files.

    """
    files = []
    for root, _, filei in os.walk(rootdir):
        for i in filei:
            if match is None:
                files.append(os.path.join(root, i))
            elif i == match:
                files.append(os.path.join(root, i))
    return files


def clean_dir(dirname):
    """Remove ALL files in the given directory."""
    for files in os.listdir(dirname):
        filename = os.path.join(dirname, files)
        if os.path.isfile(filename):
            os.remove(filename)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of running a MD simulation using the PyRETIS library.

The system considered is a simple Lennard-Jones fluid.

"""
# pylint: disable=invalid-name
import sys
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import gridspec
from pyretis.core.units import create_conversion_factors
from pyretis.setup import create_simulation
from pyretis.inout.formats import ThermoTableFormatter
from pyretis.inout.fileio import FileIO
# for plotting:
from pyretis.inout.plotting import mpl_set_style
# simulation settings:
settings = {}
settings['simulation'] = {
    'task': 'md-nve',
    'steps': 100
}
settings['system'] = {
    'units': 'lj',
    'temperature': 2.0,
    'dimensions': 3
}
settings['engine'] = {
    'class': 'velocityverlet',
    'timestep': 0.002
}
settings['output'] = {
    'backup': 'overwrite',
    'energy-file': 1,
    'screen': 10,
    'trajectory-file': 1
}
settings['potential'] = [
    {'class': 'PairLennardJonesCutnp',
     'parameter': {0: {'sigma': 1, 'epsilon': 1, 'factor': 2.5}},
     'shift': True}
]
settings['particles'] = {
    'position': {'generate': 'fcc', 'repeat': [3, 3, 3], 'density': 0.9},
    'velocity': {'generate': 'maxwell', 'momentum': True, 'seed': 0}
}

create_conversion_factors(settings['system']['units'])
simulation_nve = create_simulation(settings)

# set up extra output:
thermo_file = FileIO('thermo-test.txt', 'w', ThermoTableFormatter())
thermo_file.open()
store_results = []
# also create some other outputs:
simulation_nve.set_up_output(settings, progress=False)
# run the simulation :-)
for result in simulation_nve.run():
    stepno = result['cycle']['stepno']
    thermo_file.output(stepno, result['thermo'])
    result['thermo']['stepno'] = stepno
    store_results.append(result['thermo'])
thermo_file.close()

# We are now done with the actual simulation. Let us now do some
# simple plotting of energies:
mpl_set_style()  # load pyretis style
step = [res['stepno'] for res in store_results]
pot_e = [res['vpot'] for res in store_results]
kin_e = [res['ekin'] for res in store_results]
tot_e = [res['etot'] for res in store_results]
pressure = [res['press'] for res in store_results]
temp = [res['temp'] for res in store_results]
# first figure - some energies
fig1 = plt.figure()
gs = gridspec.GridSpec(2, 2)
ax1 = fig1.add_subplot(gs[:, 0])
ax1.plot(step, pot_e, label='Potential')
ax1.plot(step, kin_e, label='Kinetic')
ax1.plot(step, tot_e, label='Total')
ax1.set_xlabel('Step no.')
ax1.set_ylabel('Energy per particle')
ax1.legend(loc='center left', prop={'size': 'small'})

ax2 = fig1.add_subplot(gs[0, 1])
ax2.plot(step, temp)
ax2.set_ylabel('Temperature')

ax3 = fig1.add_subplot(gs[1, 1])
ax3.plot(step, pressure)
ax3.set_xlabel('Step no.')
ax3.set_ylabel('Pressure')

fig1.subplots_adjust(bottom=0.12, right=0.95, top=0.95, left=0.12, wspace=0.3)
# second figure, momentum in different directions
momentum = np.array([res['mom'] for res in store_results])
fig2 = plt.figure()
ax4 = fig2.add_subplot(111)
ax4.plot(step, momentum[:, 0], lw=4, alpha=0.7, label='x')
ax4.plot(step, momentum[:, 1], lw=4, alpha=0.7, label='y')
ax4.plot(step, momentum[:, 2], lw=4, alpha=0.7, label='z')
ax4.set_xlabel('Step')
ax4.set_ylabel('Linear momentum')
ax4.legend(loc='upper center', prop={'size': 'small'}, ncol=3)
if 'noplot' not in sys.argv[1:]:
    plt.show()
